"""
Module for the source catalog step.
"""

import numpy as np
from astropy.table import Table
from roman_datamodels import datamodels, maker_utils
from roman_datamodels.datamodels import ImageModel, MosaicModel
from roman_datamodels.maker_utils import mk_datamodel

from romancal.source_catalog.background import RomanBackground
from romancal.source_catalog.detection import convolve_data, make_segmentation_image
from romancal.source_catalog.reference_data import ReferenceData
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.stpipe import RomanStep

__all__ = ["SourceCatalogStep"]


class SourceCatalogStep(RomanStep):
    """
    Create a catalog of sources including photometry and basic shape
    measurements.

    Parameters
    -----------
    input : str, `ImageModel`, or `MosaicModel`
        Path to an ASDF file, or an `ImageModel` or `MosaicModel`.
    """

    class_alias = "source_catalog"
    reference_file_types = []

    spec = """
        bkg_boxsize = integer(default=1000)   # background mesh box size in pixels
        kernel_fwhm = float(default=2.0)      # Gaussian kernel FWHM in pixels
        snr_threshold = float(default=3.0)    # per-pixel SNR threshold above the bkg
        npixels = integer(default=25)         # min number of pixels in source
        deblend = boolean(default=False)      # deblend sources?
        aperture_ee1 = integer(default=30)    # aperture encircled energy 1
        aperture_ee2 = integer(default=50)    # aperture encircled energy 2
        aperture_ee3 = integer(default=70)    # aperture encircled energy 3
        ci1_star_threshold = float(default=2.0)  # CI 1 star threshold
        ci2_star_threshold = float(default=1.8)  # CI 2 star threshold
        suffix = string(default='cat')        # Default suffix for output files
        fit_psf = boolean(default=True)      # fit source PSFs for accurate astrometry?
    """

    def process(self, step_input):
        if isinstance(step_input, datamodels.DataModel):
            input_model = step_input
        else:
            input_model = datamodels.open(step_input)

        if not isinstance(input_model, (ImageModel, MosaicModel)):
            raise ValueError("The input model must be an ImageModel or MosaicModel.")

        # Copy the data and error arrays to avoid modifying the input
        # model. We use mk_datamodel to copy *only* the data and err
        # arrays. The metadata and dq and weight arrays are not copied
        # because they are not modified in this step. The other model
        # arrays (e.g., var_rnoise) are not currently used by this step.
        if isinstance(input_model, ImageModel):
            model = mk_datamodel(
                ImageModel,
                meta=input_model.meta,
                shape=(0, 0),
                data=input_model.data.copy(),
                err=input_model.err.copy(),
                dq=input_model.dq,
            )
        elif isinstance(input_model, MosaicModel):
            model = mk_datamodel(
                MosaicModel,
                meta=input_model.meta,
                shape=(0, 0),
                data=input_model.data.copy(),
                err=input_model.err.copy(),
                weight=input_model.weight,
            )

        if isinstance(model, ImageModel):
            cat_model = datamodels.SourceCatalogModel
        else:
            cat_model = datamodels.MosaicSourceCatalogModel
        source_catalog_model = maker_utils.mk_datamodel(cat_model)

        for key in source_catalog_model.meta.keys():
            value = (
                model.meta.instrument[key]
                if key == "optical_element"
                else model.meta[key]
            )
            source_catalog_model.meta[key] = value
        aperture_ee = (self.aperture_ee1, self.aperture_ee2, self.aperture_ee3)
        refdata = ReferenceData(model, aperture_ee)
        aperture_params = refdata.aperture_params

        mask = np.isnan(model.data)
        coverage_mask = np.isnan(model.err)
        bkg = RomanBackground(
            model.data,
            box_size=self.bkg_boxsize,
            mask=mask,
            coverage_mask=coverage_mask,
        )
        model.data -= bkg.background

        convolved_data = convolve_data(
            model.data, kernel_fwhm=self.kernel_fwhm, mask=coverage_mask
        )

        segment_img = make_segmentation_image(
            convolved_data,
            snr_threshold=self.snr_threshold,
            npixels=self.npixels,
            bkg_rms=bkg.background_rms,
            deblend=self.deblend,
            mask=coverage_mask,
        )

        if segment_img is None:  # no sources found
            source_catalog_model.source_catalog = Table()
        else:
            ci_star_thresholds = (
                self.ci1_star_threshold,
                self.ci2_star_threshold,
            )
            catobj = RomanSourceCatalog(
                model,
                segment_img,
                convolved_data,
                aperture_params,
                ci_star_thresholds,
                self.kernel_fwhm,
                self.fit_psf,
            )

            # put the resulting catalog in the model
            source_catalog_model.source_catalog = catobj.catalog

        # always save the segmentation image and source catalog
        self.save_base_results(segment_img, source_catalog_model)

        # Return the source catalog object or the input model. If the
        # input model is an ImageModel, the metadata is updated with the
        # source catalog filename.
        if getattr(self, "return_updated_model", False):
            # define the catalog filename; self.save_model will
            # determine whether to use a fully qualified path
            output_catalog_name = self.make_output_path(
                basepath=model.meta.filename, suffix="cat"
            )

            # set the suffix to something else to prevent the step from
            # overwriting the source catalog file with a datamodel
            self.suffix = "sourcecatalog"

            if isinstance(input_model, ImageModel):
                update_metadata(input_model, output_catalog_name)

            result = input_model
        else:
            result = source_catalog_model

        return result

    def save_base_results(self, segment_img, source_catalog_model):
        # save the segmentation map and source catalog
        output_filename = (
            self.output_file
            if self.output_file is not None
            else source_catalog_model.meta.filename
        )

        if isinstance(source_catalog_model, datamodels.SourceCatalogModel):
            seg_model = datamodels.SegmentationMapModel
        else:
            seg_model = datamodels.MosaicSegmentationMapModel

        segmentation_model = maker_utils.mk_datamodel(seg_model)
        for key in segmentation_model.meta.keys():
            segmentation_model.meta[key] = source_catalog_model.meta[key]

        if segment_img is not None:
            segmentation_model.data = segment_img.data.astype(np.uint32)
            self.save_model(
                segmentation_model,
                output_file=output_filename,
                suffix="segm",
                force=True,
            )

        # save the source catalog
        self.save_model(
            source_catalog_model,
            output_file=output_filename,
            suffix="cat",
            force=True,
        )


def update_metadata(model, output_catalog_name):
    # update datamodel to point to the source catalog file destination
    model.meta["source_detection"] = maker_utils.mk_source_detection(
        tweakreg_catalog_name=output_catalog_name
    )
    model.meta.cal_step.source_detection = "COMPLETE"



"""
This module collects all of the stpipe.Pipeline subclasses
made available by this package.
"""

from .exposure_pipeline import ExposurePipeline
from .mosaic_pipeline import MosaicPipeline

__all__ = ["ExposurePipeline", "MosaicPipeline"]



""" Roman tests for generating associatinos based on skycells"""

import os

import pytest

from romancal.associations import skycell_asn


@pytest.mark.bigdata
def test_skycell_asn_generation(rtdata):
    """Test for the generation of associations based on skycells"""

    # This test should generate seven json files
    args = [
        "r0000101001001001001_0002_wfi01_cal.asdf",
        "r0000101001001001001_0002_wfi10_cal.asdf",
        "-o",
        "r512",
    ]
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi01_cal.asdf")
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi10_cal.asdf")

    skycell_asn.Main(args)

    # skycell associations that should be generated
    output_files = [
        "r512_p_v01001001001_r274dp63x31y80_f158_coadd_asn.json",
        "r512_p_v01001001001_r274dp63x31y81_f158_coadd_asn.json",
        "r512_p_v01001001001_r274dp63x32y82_f158_coadd_asn.json",
        "r512_p_v01001001001_r274dp63x32y80_f158_coadd_asn.json",
        "r512_p_v01001001001_r274dp63x32y81_f158_coadd_asn.json",
        "r512_p_v01001001001_r274dp63x32y82_f158_coadd_asn.json",
    ]
    # Test that the json files exist
    for file in output_files:
        skycell_asn.logger.info(f"Check that the json file exists{file}")
        assert os.path.isfile(file)



"""Regression test for the linearity correction step."""

import pytest

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_linearity_step(rtdata, ignore_asdf_paths):
    """Function to run and compare linearity correction files."""
    input_file = "r0000101001001001001_0001_wfi01_refpix.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    args = ["romancal.step.LinearityStep", rtdata.input]
    RomanStep.from_cmdline(args)
    output = "r0000101001001001001_0001_wfi01_linearity.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
def test_linearity_outfile_step(rtdata, ignore_asdf_paths):
    """Function to run and linearity correction files. Here the
    test is for renaming the output file."""
    input_file = "r0000101001001001001_0001_wfi01_refpix.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    args = ["romancal.step.LinearityStep", rtdata.input, "--output_file=Test_linearity"]
    RomanStep.from_cmdline(args)
    output = "Test_linearity.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()



"""Regression test for the Dark current subtraction step."""

import pytest

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_dark_current_subtraction_step(rtdata, ignore_asdf_paths):
    """Function to run and compare Dark Current subtraction files. Note: This
    should include tests for overrides etc."""

    input_datafile = "r0000101001001001001_0001_wfi01_linearity.asdf"
    rtdata.get_data(f"WFI/image/{input_datafile}")
    rtdata.input = input_datafile

    args = ["romancal.step.DarkCurrentStep", rtdata.input]
    RomanStep.from_cmdline(args)
    output = "r0000101001001001001_0001_wfi01_darkcurrent.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
def test_dark_current_outfile_step(rtdata, ignore_asdf_paths):
    """Function to run and compare Dark Current subtraction files. Here the
    test is for renaming the output file."""
    input_datafile = "r0000101001001001001_0001_wfi01_linearity.asdf"
    rtdata.get_data(f"WFI/image/{input_datafile}")
    rtdata.input = input_datafile

    args = [
        "romancal.step.DarkCurrentStep",
        rtdata.input,
        "--output_file=Test_darkcurrent",
    ]
    RomanStep.from_cmdline(args)
    output = "Test_darkcurrent.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
def test_dark_current_outfile_suffix(rtdata, ignore_asdf_paths):
    """Function to run and compare Dark Current subtraction files. Here the
    test is for renaming the output file."""
    input_datafile = "r0000101001001001001_0001_wfi01_linearity.asdf"
    rtdata.get_data(f"WFI/image/{input_datafile}")
    rtdata.input = input_datafile

    args = [
        "romancal.step.DarkCurrentStep",
        rtdata.input,
        "--output_file=Test_dark",
        '--suffix="suffix_test"',
    ]
    RomanStep.from_cmdline(args)
    output = "Test_darkcurrent.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
def test_dark_current_output(rtdata, ignore_asdf_paths):
    """Function to run and compare Dark Current subtraction files. Here the
    test for overriding the CRDS dark reference file."""

    input_datafile = "r0000101001001001001_0001_wfi01_linearity.asdf"
    rtdata.get_data(f"WFI/image/{input_datafile}")
    rtdata.input = input_datafile
    dark_output_name = "r0000101001001001001_0001_wfi01_darkcurrent.asdf"

    args = [
        "romancal.step.DarkCurrentStep",
        rtdata.input,
        f"--dark_output={dark_output_name}",
    ]
    RomanStep.from_cmdline(args)
    output = "r0000101001001001001_0001_wfi01_darkcurrent.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()



"""Public common step definition for OutlierDetection processing."""

from functools import partial

from romancal.datamodels import ModelLibrary
from romancal.outlier_detection.utils import detect_outliers

from ..stpipe import RomanStep

__all__ = ["OutlierDetectionStep"]


class OutlierDetectionStep(RomanStep):
    """Flag outlier bad pixels and cosmic rays in DQ array of each input image.

    Input images can be listed in an input association file or already wrapped
    with a ModelLibrary. DQ arrays are modified in place.

    Parameters
    -----------
    input_data : `~romancal.datamodels.container.ModelLibrary`
        A `~romancal.datamodels.container.ModelLibrary` object.

    """

    class_alias = "outlier_detection"

    # The members of spec needs to be a super-set of all parameters needed
    # by the various versions of the outlier_detection algorithms, and each
    # version will pick and choose what they need while ignoring the rest.
    spec = """
        weight_type = option('ivm','exptime',default='ivm') # Weighting type to use to create the median image
        pixfrac = float(default=1.0) # Fraction by which input pixels are shrunk before being drizzled onto the output image grid
        kernel = string(default='square') # Shape of the kernel used for flux distribution onto output images
        fillval = string(default='INDEF') # Value assigned to output pixels that have zero weight or no flux during drizzling
        maskpt = float(default=0.7) # Percentage of weight image values below which they are flagged as bad pixels
        snr = string(default='5.0 4.0') # The signal-to-noise values to use for bad pixel identification
        scale = string(default='1.2 0.7') # The scaling factor applied to derivative used to identify bad pixels
        backg = float(default=0.0) # User-specified background value to subtract during final identification step
        save_intermediate_results = boolean(default=False) # Specifies whether or not to write out intermediate products to disk
        resample_data = boolean(default=True) # Specifies whether or not to resample the input images when performing outlier detection
        good_bits = string(default="~DO_NOT_USE+NON_SCIENCE")  # DQ bit value to be considered 'good'
        in_memory = boolean(default=False) # Specifies whether or not to keep all intermediate products and datamodels in memory
    """  # noqa: E501

    def process(self, input_models):
        """Perform outlier detection processing on input data."""

        if isinstance(input_models, ModelLibrary):
            library = input_models
        else:
            library = ModelLibrary(input_models)

        # check number of input models
        if len(library) < 2:
            # if input can be parsed into a ModelLibrary
            # but is not valid then log a warning message and
            # skip outlier detection step
            self.log.warning(
                "Skipping outlier_detection - at least two imaging observations are needed."
            )

            def set_skip(model, index):
                model.meta.cal_step["outlier_detection"] = "SKIPPED"

            list(library.map_function(set_skip))

            return library

        # check that all inputs are WFI_IMAGE
        def get_exptype(model, index):
            return model.meta.exposure.type

        exptypes = list(library.map_function(get_exptype, modify=False))
        if any(exptype != "WFI_IMAGE" for exptype in exptypes):
            raise ValueError(
                f"outlier_detection only supports WFI_IMAGE exposure types: {set(exptypes)}"
            )

        # Setup output path naming if associations are involved.
        asn_id = library.asn.get("asn_id", None)
        if asn_id is not None:
            _make_output_path = self.search_attr("_make_output_path", parent_first=True)
            self._make_output_path = partial(_make_output_path, asn_id=asn_id)

        snr1, snr2 = (float(v) for v in self.snr.split())
        scale1, scale2 = (float(v) for v in self.scale.split())

        # Set up outlier detection, then do detection
        detect_outliers(
            library,
            self.weight_type,
            self.pixfrac,
            self.kernel,
            self.fillval,
            self.maskpt,
            snr1,
            snr2,
            scale1,
            scale2,
            self.backg,
            self.save_intermediate_results,
            self.resample_data,
            self.good_bits,
            self.in_memory,
            self.make_output_path,
        )
        return library



"""

Unit tests for linearity correction

"""

import numpy as np
import pytest
from roman_datamodels import dqflags, maker_utils
from roman_datamodels.datamodels import LinearityRefModel, ScienceRawModel

from romancal.dq_init import DQInitStep
from romancal.linearity import LinearityStep


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_linearity_coeff(instrument, exptype):
    """Test that the basic inferface works for data requiring a DQ reffile"""

    # Set test size
    shape = (5, 20, 20)

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape)
    wfi_sci_raw.meta.instrument.name = instrument
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guide_star"]["window_xstart"] = 1012
    wfi_sci_raw.meta["guide_star"]["window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = exptype
    wfi_sci_raw.data = np.ones(shape, dtype=np.uint16)
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw)

    result = DQInitStep.call(wfi_sci_raw_model)
    result = LinearityStep.call(result, override_linearity="N/A")

    assert result.meta.cal_step.linearity == "SKIPPED"

    # Set coefficient values in reference file to check the algorithm
    # Equation is DNcorr = L0 + L1*DN(i) + L2*DN(i)^2 + L3*DN(i)^3 + L4*DN(i)^4
    # DN(i) = signal in pixel, Ln = coefficient from ref file
    # L0 = 0 for all pixels for CDP6

    coeffs = np.zeros(shape, dtype=np.float32)
    coeffs[1, :, :].fill(0.85)
    coeffs[2, :, :].fill(4.62e-06)
    coeffs[3, :, :].fill(-6.16e-11)
    coeffs[4, :, :].fill(7.23e-16)
    # coeffs = np.tile(coeffs,(5,20,4))
    # set one set of coeffs to 0 to check that the data values are not altered.
    # and that the DQ flag is set to NO_LIN_CORR
    coeffs[0:, 5, 5] = 0.0
    # set one of the coeffs to NaN and check that the DQ flag is set to NO_LIN_CORR
    coeffs[0:, 6, 5] = np.nan
    # save the pixel values to make sure they are not altered
    pixels_55 = result.data[0:, 5, 5]
    pixels_65 = result.data[0:, 6, 5]
    linref = maker_utils.mk_linearity(shape=shape, coeffs=coeffs)
    linref_model = LinearityRefModel(linref)

    LinearityStep.call(result, override_linearity=linref_model)

    assert result.meta.cal_step.linearity == "COMPLETE"
    assert result.pixeldq[5, 5] == dqflags.pixel["NO_LIN_CORR"]
    assert result.pixeldq[6, 5] == dqflags.pixel["NO_LIN_CORR"]
    np.testing.assert_array_equal(result.data[0:, 5, 5], pixels_55)
    np.testing.assert_array_equal(result.data[0:, 6, 5], pixels_65)



import numpy as np
from roman_datamodels.datamodels import ImageModel, MosaicModel

from romancal.datamodels import ModelLibrary
from romancal.source_catalog.background import RomanBackground


def subtract_background(model, box_size=1000):
    """
    Subtract the background from the input model.

    Parameters
    ----------
    model : ImageModel or MosaicModel
        The input model to subtract the background from.

    box_size : int
        The size of the box to use for the background estimation.

    Returns
    -------
    model : ImageModel or MosaicModel
        The input model with the background subtracted.
    """
    if not isinstance(model, (ImageModel, MosaicModel)):
        raise ValueError("The input model must be an ImageModel or MosaicModel.")

    # Subtract the background
    mask = np.isnan(model.data)
    coverage_mask = np.logical_or(np.isnan(model.err), model.err == 0)
    bkg = RomanBackground(
        model.data,
        box_size=box_size,
        mask=mask,
        coverage_mask=coverage_mask,
    )
    model.data -= bkg.background
    return model


def subtract_background_library(library, box_size=1000):
    """
    Subtract the background from all models in the input library.

    Parameters
    ----------
    library : ModelLibrary
        The input library to subtract the background from.

    box_size : int
        The size of the box to use for the background estimation.

    Returns
    -------
    library : ModelLibrary
        The input library with the background subtracted.
    """
    if not isinstance(library, ModelLibrary):
        raise TypeError("library input must be a ModelLibrary object")

    with library:
        for model in library:
            model = subtract_background(model, box_size=box_size)
            library.shelve(model)

    return library



import logging
import os
from copy import deepcopy

import asdf
import numpy as np
from astropy.extern.configobj.configobj import ConfigObj
from astropy.extern.configobj.validate import Validator
from roman_datamodels import datamodels
from stcal.alignment import util

from ..datamodels import ModelLibrary
from ..stpipe import RomanStep
from . import resample

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["ResampleStep"]

# conversion factor from steradian to squared arcsec
SR_TO_ARCSEC2 = 4.254517e10


class ResampleStep(RomanStep):
    """
    Resample input data onto a regular grid using the drizzle algorithm.

    .. note::
        When supplied via ``output_wcs``, a custom WCS overrides other custom
        WCS parameters such as ``output_shape`` (now computed from by
        ``output_wcs.bounding_box``), ``crpix``

    Parameters
    -----------
    input : str, `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.ModelLibrary`
        If a string is provided, it should correspond to either a single ASDF filename
        or an association filename. Alternatively, a single DataModel instance can be
        provided instead of an ASDF filename. Multiple files can be processed via
        either an association file or wrapped by a
        `~romancal.datamodels.ModelLibrary`.

    Returns
    -------
    : `roman_datamodels.datamodels.MosaicModel`
        A mosaic datamodel with the final output frame.
    """  # noqa: E501

    class_alias = "resample"

    spec = """
        pixfrac = float(default=1.0) # change back to None when drizpar reference files are updated
        kernel = string(default='square') # change back to None when drizpar reference files are updated
        fillval = string(default='INDEF' ) # change back to None when drizpar reference files are updated
        weight_type = option('ivm', 'exptime', None, default='ivm')  # change back to None when drizpar ref update
        output_shape = int_list(min=2, max=2, default=None)  # [x, y] order
        crpix = float_list(min=2, max=2, default=None)
        crval = float_list(min=2, max=2, default=None)
        rotation = float(default=None)
        pixel_scale_ratio = float(default=1.0) # Ratio of input to output pixel scale
        pixel_scale = float(default=None) # Absolute pixel scale in arcsec
        output_wcs = string(default='')  # Custom output WCS.
        single = boolean(default=False)
        blendheaders = boolean(default=True)
        allowed_memory = float(default=None)  # Fraction of memory to use for the combined image.
        in_memory = boolean(default=True)
        good_bits = string(default='~DO_NOT_USE+NON_SCIENCE')  # The good bits to use for building the resampling mask.
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):
        if isinstance(input, datamodels.DataModel):
            input_models = ModelLibrary([input])
            # set output filename from meta.filename found in the first datamodel
            output = input.meta.filename
            self.blendheaders = False
        elif isinstance(input, str):
            # either a single asdf filename or an association filename
            try:
                # association filename
                input_models = ModelLibrary(input)
            except Exception:
                # single ASDF filename
                input_models = ModelLibrary([input])
            output = input_models.asn["products"][0]["name"]
        elif isinstance(input, ModelLibrary):
            input_models = input
            if "name" in input_models.asn["products"][0]:
                output = input_models.asn["products"][0]["name"]
            else:
                # set output filename using the common prefix of all datamodels
                output = f"{os.path.commonprefix([x['expname'] for x in input_models.asn['products'][0]['members']])}.asdf"
                if len(output) == 0:
                    # set default filename if no common prefix can be determined
                    output = "resample_output.asdf"
        else:
            raise TypeError(
                "Input must be an ASN filename, a ModelLibrary, "
                "a single ASDF filename, or a single Roman DataModel."
            )

        # Check that input models are 2D images
        with input_models:
            example_model = input_models.borrow(0)
            data_shape = example_model.data.shape
            input_models.shelve(example_model, 0, modify=False)
            if len(data_shape) != 2:
                # resample can only handle 2D images, not 3D cubes, etc
                raise RuntimeError(f"Input {input_models[0]} is not a 2D image.")

        self.wht_type = self.weight_type
        self.log.info("Setting drizzle's default parameters...")
        kwargs = self.set_drizzle_defaults()
        kwargs["allowed_memory"] = self.allowed_memory

        # Issue a warning about the use of exptime weighting
        if self.wht_type == "exptime":
            self.log.warning("Use of EXPTIME weighting will result in incorrect")
            self.log.warning("propagated errors in the resampled product")

        # Custom output WCS parameters.
        # Modify get_drizpars if any of these get into reference files:
        kwargs["output_shape"] = self._check_list_pars(
            self.output_shape, "output_shape", min_vals=[1, 1]
        )
        kwargs["output_wcs"] = self._load_custom_wcs(
            self.output_wcs, kwargs["output_shape"]
        )
        kwargs["crpix"] = self._check_list_pars(self.crpix, "crpix")
        kwargs["crval"] = self._check_list_pars(self.crval, "crval")
        kwargs["rotation"] = self.rotation
        kwargs["pscale"] = self.pixel_scale
        kwargs["pscale_ratio"] = self.pixel_scale_ratio
        kwargs["in_memory"] = self.in_memory

        # Call the resampling routine
        resamp = resample.ResampleData(input_models, output=output, **kwargs)
        result = resamp.do_drizzle()

        with result:
            for i, model in enumerate(result):
                self._final_updates(model, input_models, kwargs)
                result.shelve(model, i)
            if len(result) == 1:
                model = result.borrow(0)
                result.shelve(model, 0, modify=False)
                return model

        return result

    def _final_updates(self, model, input_models, kwargs):
        model.meta.cal_step["resample"] = "COMPLETE"
        model.meta.wcsinfo.s_region = util.compute_s_region_imaging(
            model.meta.wcs, model.data.shape
        )

        # if pixel_scale exists, it will override pixel_scale_ratio.
        # calculate the actual value of pixel_scale_ratio based on pixel_scale
        # because source_catalog uses this value from the header.
        model.meta.resample.pixel_scale_ratio = (
            self.pixel_scale / np.sqrt(model.meta.photometry.pixel_area * SR_TO_ARCSEC2)
            if self.pixel_scale
            else self.pixel_scale_ratio
        )
        model.meta.resample.pixfrac = kwargs["pixfrac"]
        self.update_phot_keywords(model)
        model.meta.resample["good_bits"] = kwargs["good_bits"]

    @staticmethod
    def _check_list_pars(vals, name, min_vals=None):
        """
        Check if a specific keyword parameter is properly formatted.

        Parameters
        ----------
        vals : list or tuple
            A list or tuple containing a pair of values currently assigned to the
            keyword parameter `name`. Both values must be either `None` or not `None`.
        name : str
            The name of the keyword parameter.
        min_vals : list or tuple, optional
            A list or tuple containing a pair of minimum values to be assigned
            to `name`, by default None.

        Returns
        -------
        None or list
            If either `vals` is set to `None` (or both of its elements), the
            returned result will be `None`. Otherwise, the returned result will be
            a list containing the current values assigned to `name`.

        Raises
        ------
        ValueError
            This error will be raised if any of the following conditions are met:
            - the number of elements of `vals` is not 2;
            - the currently assigned values of `vals` are smaller than the
            minimum value provided;
            - one element is `None` and the other is not `None`.
        """
        if vals is None:
            return None
        if len(vals) != 2:
            raise ValueError(f"List '{name}' must have exactly two elements.")
        n = sum(x is None for x in vals)
        if n == 2:
            return None
        elif n == 0:
            if min_vals and sum(x >= y for x, y in zip(vals, min_vals)) != 2:
                raise ValueError(
                    f"'{name}' values must be larger or equal to {list(min_vals)}"
                )
            return list(vals)
        else:
            raise ValueError(f"Both '{name}' values must be either None or not None.")

    @staticmethod
    def _load_custom_wcs(asdf_wcs_file, output_shape):
        if not asdf_wcs_file:
            return None

        with asdf.open(asdf_wcs_file) as af:
            wcs = deepcopy(af.tree["wcs"])

        if output_shape is not None:
            wcs.array_shape = output_shape[::-1]
        elif wcs.pixel_shape is not None:
            wcs.array_shape = wcs.pixel_shape[::-1]
        elif wcs.bounding_box is not None:
            wcs.array_shape = tuple(
                int(axs[1] - axs[0] + 0.5)
                for axs in wcs.bounding_box.bounding_box(order="C")
            )
        elif wcs.array_shape is None:
            raise ValueError(
                "Step argument 'output_shape' is required when custom WCS "
                "does not have neither of 'array_shape', 'pixel_shape', or "
                "'bounding_box' attributes set."
            )

        return wcs

    def update_phot_keywords(self, model):
        """Update pixel scale keywords"""
        if model.meta.photometry.pixel_area is not None:
            model.meta.photometry.pixel_area *= model.meta.resample.pixel_scale_ratio**2

    def set_drizzle_defaults(self):
        """Set the default parameters for drizzle."""
        configspec = self.load_spec_file()
        config = ConfigObj(configspec=configspec)
        if config.validate(Validator()):
            kwargs = config.dict()

        if self.pixfrac is None:
            self.pixfrac = 1.0
        if self.kernel is None:
            self.kernel = "square"
        if self.fillval is None:
            self.fillval = "INDEF"
        # Force definition of good bits
        kwargs["good_bits"] = self.good_bits
        kwargs["pixfrac"] = self.pixfrac
        kwargs["kernel"] = str(self.kernel)
        kwargs["fillval"] = str(self.fillval)
        #  self.weight_type has a default value of None
        # The other instruments read this parameter from a reference file
        if self.wht_type is None:
            self.wht_type = "ivm"

        kwargs["wht_type"] = str(self.wht_type)
        kwargs["pscale_ratio"] = self.pixel_scale_ratio
        kwargs.pop("pixel_scale_ratio")

        for k, v in kwargs.items():
            if k in [
                "pixfrac",
                "kernel",
                "fillval",
                "wht_type",
                "pscale_ratio",
            ]:
                log.info("  using: %s=%s", k, repr(v))

        return kwargs



"""Association attributes common to DMS-based Rules"""

from romancal.associations.exceptions import AssociationNotValidError
from romancal.associations.lib.acid import ACIDMixin
from romancal.associations.lib.constraint import (
    AttrConstraint,
    Constraint,
    SimpleConstraint,
)
from romancal.associations.lib.counter import Counter
from romancal.associations.lib.utilities import getattr_from_list

__all__ = ["Constraint_TargetAcq", "Constraint_WFSC", "DMSBaseMixin"]

# Default product name
PRODUCT_NAME_DEFAULT = "undefined"

# DMS file name templates
_ASN_NAME_TEMPLATE_STAMP = "r{program}-{acid}_{stamp}_{type}_{sequence:03d}_asn"
_ASN_NAME_TEMPLATE = "r{program}-{acid}_{type}_{sequence:03d}_asn"

# Acquisition and Confirmation images
ACQ_EXP_TYPES = (
    "nrc_taconfirm",
    "nrc_tacq",
)

# Exposure EXP_TYPE to Association EXPTYPE mapping
# flake8: noqa: E241
EXPTYPE_MAP = {
    "nrc_dark": "dark",
    "nrc_flat": "flat",
    "nrc_focus": "engineering",
    "nrc_led": "engineering",
    "nrc_tacq": "target_acquisition",
    "nrc_taconfirm": "target_acquisition",
}

# Coronographic exposures
CORON_EXP_TYPES = ["mir_4qpm", "mir_lyot", "nrc_coron"]

# Roman WFI detectors
WFI_DETECTORS = [
    "wfi01",
    "wfi02",
    "wfi03",
    "wfi04",
    "wfi05",
    "wfi06",
    "wfi07",
    "wfi08",
    "wfi09",
    "wfi10",
    "wfi11",
    "wfi12",
    "wfi13",
    "wfi14",
    "wfi15",
    "wfi16",
    "wfi17",
    "wfi18",
]

# Exposures that get Level2b processing
IMAGE2_SCIENCE_EXP_TYPES = [
    "wfi_image",
]

IMAGE2_NONSCIENCE_EXP_TYPES = [
    "wfi_focus",
]
IMAGE2_NONSCIENCE_EXP_TYPES.extend(ACQ_EXP_TYPES)

SPEC2_SCIENCE_EXP_TYPES = [
    "wfi_grism",
    "wfi_prism",
]

SPECIAL_EXPOSURE_MODIFIERS = {
    "background": ["bkgdtarg"],
    "imprint": ["is_imprt"],
    "psf": ["is_psf"],
}

#
# Key that uniquely identifies members.
MEMBER_KEY = "expname"

# Non-specified values found in DMS Association Pools
_EMPTY = (
    None,
    "",
    "NULL",
    "Null",
    "null",
    "--",
    "N",
    "n",
    "F",
    "f",
    "FALSE",
    "false",
    "False",
    "N/A",
    "n/a",
)

# Degraded status information
_DEGRADED_STATUS_OK = "No known degraded exposures in association."
_DEGRADED_STATUS_NOTOK = (
    "One or more members have an error associated with them.\nDetails can be found in"
    " the member.exposerr attribute."
)


class DMSBaseMixin(ACIDMixin):
    """Association attributes common to DMS-based Rules

    Attributes
    ----------
    sequence : int
        The sequence number of the current association
    """

    # Associations of the same type are sequenced.
    _sequence = Counter(start=1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._acid = None
        self._asn_name = None
        self.sequence = None
        if "degraded_status" not in self.data:
            self.data["degraded_status"] = _DEGRADED_STATUS_OK
        if "program" not in self.data:
            self.data["program"] = "noprogram"

    @classmethod
    def create(cls, item, version_id=None):
        """Create association if item belongs

        Parameters
        ----------
        item : dict
            The item to initialize the association with.

        version_id : str or None
            Version_Id to use in the name of this association.
            If None, nothing is added.

        Returns
        -------
        (association, reprocess_list)
            2-tuple consisting of:

                - association : The association or, if the item does not
                  match this rule, None
                - [ProcessList[, ...]]: List of items to process again.
        """
        asn, reprocess = super().create(item, version_id)
        if not asn:
            return None, reprocess
        asn.sequence = next(asn._sequence)
        return asn, reprocess

    @property
    def acid(self):
        """Association ID"""
        acid = self._acid
        if self._acid is None:
            acid = self.acid_from_constraints()
        return acid

    @property
    def asn_name(self):
        """The association name

        The name that identifies this association. When dumped,
        will form the basis for the suggested file name.

        Typically, it is generated based on the current state of
        the association, but can be overridden.
        """
        if self._asn_name:
            return self._asn_name

        program = self.data["program"]
        version_id = self.version_id
        asn_type = self.data["asn_type"]
        sequence = self.sequence
        target = self.target

        if version_id:
            name = _ASN_NAME_TEMPLATE_STAMP.format(
                program=program,
                acid=self.acid.id,
                stamp=version_id,
                type=asn_type,
                sequence=sequence,
                target=target,
            )
        else:
            name = _ASN_NAME_TEMPLATE.format(
                program=program,
                acid=self.acid.id,
                type=asn_type,
                sequence=sequence,
                target=target,
            )
        return name.lower()

    @asn_name.setter
    def asn_name(self, name):
        """Override calculated association name"""
        self._asn_name = name

    @property
    def current_product(self):
        """Return currnet products"""
        return self.data["products"][-1]

    @property
    def from_items(self):
        """The list of items that contributed to the association."""
        try:
            items = [
                member.item
                for product in self["products"]
                for member in product["members"]
            ]
        except KeyError:
            items = []
        return items

    @property
    def member_ids(self):
        """Set of all member ids in all products of this association"""
        member_ids = {
            member[MEMBER_KEY]
            for product in self["products"]
            for member in product["members"]
        }
        return member_ids

    @property
    def validity(self):
        """Keeper of the validity tests"""
        try:
            validity = self._validity
        except AttributeError:
            self._validity = {}
            validity = self._validity
        return validity

    @validity.setter
    def validity(self, item):
        """Set validity dict"""
        self._validity = item

    def get_exposure_type(self, item, default="science"):
        """Determine the exposure type of a pool item

        Parameters
        ----------
        item : dict
            The pool entry to determine the exposure type of

        default : str or None
            The default exposure type.
            If None, routine will raise LookupError

        Returns
        -------
        exposure_type : str
            Exposure type. Can be one of

                - 'science': Item contains science data
                - 'target_acquisition': Item contains target acquisition data.
                - 'autoflat': NIRSpec AUTOFLAT
                - 'autowave': NIRSpec AUTOWAVE
                - 'psf': PSF
                - 'imprint': MSA/IFU Imprint/Leakcal

        Raises
        ------
        LookupError
            When `default` is None and an exposure type cannot be determined
        """
        return get_exposure_type(item, default=default, association=self)

    def is_member(self, new_member):
        """Check if member is already a member

        Parameters
        ----------
        new_member : Member
            The member to check for
        """
        try:
            current_members = self.current_product["members"]
        except KeyError:
            return False

        for member in current_members:
            if member == new_member:
                return True
        return False

    def is_item_member(self, item):
        """Check if item is already a member of this association

        Parameters
        ----------
        item : dict
            The item to check for.

        Returns
        -------
        is_item_member : bool
            True if item is a member.
        """
        return item in self.from_items

    def item_getattr(self, item, attributes):
        """Return value from any of a list of attributes

        Parameters
        ----------
        item : dict
            item to retrieve from

        attributes : list
            List of attributes

        Returns
        -------
        (attribute, value)
            Returns the value and the attribute from
            which the value was taken.

        Raises
        ------
        KeyError
            None of the attributes are found in the dict.
        """
        return item_getattr(item, attributes, self)

    def new_product(self, product_name=PRODUCT_NAME_DEFAULT):
        """Start a new product"""
        product = {"name": product_name, "members": []}
        try:
            self.data["products"].append(product)
        except (AttributeError, KeyError):
            self.data["products"] = [product]

    def update_asn(self, item=None, member=None):
        """Update association meta information

        Parameters
        ----------
        item : dict or None
            Item to use as a source. If not given, item-specific
            information will be left unchanged.

        member : Member or None
            An association member to use as source.
            If not given, member-specific information will be update
            from current association/product membership.

        Notes
        -----
        If both `item` and `member` are given,
        information in `member` will take precedence.
        """
        self.update_degraded_status()
        self.set_visit_id(item)

    def set_visit_id(self, item):
        """Set the visit id in the association"""

        if "visit_id" in self.data:
            pass
        else:
            self.data["visit_id"] = item["visit_id"]
            try:
                self.data["visit_id"] = item["visit_id"]
            except KeyError:
                # logger.debug(f'Visit_id not found')
                self.data["visit_id"] = "None"

    def update_degraded_status(self):
        """Update association degraded status"""

        if self.data["degraded_status"] == _DEGRADED_STATUS_OK:
            for product in self.data["products"]:
                for member in product["members"]:
                    try:
                        exposerr = member["exposerr"]
                    except KeyError:
                        continue
                    else:
                        if exposerr not in _EMPTY:
                            self.data["degraded_status"] = _DEGRADED_STATUS_NOTOK
                            break

    def update_validity(self, entry):
        """Check/Update the validity of the association"""
        for test in self.validity.values():
            if not test["validated"]:
                test["validated"] = test["check"](entry)

    @classmethod
    def reset_sequence(cls):
        """Reset the sequence counter to 1"""
        cls._sequence = Counter(start=1)

    @classmethod
    def validate(cls, asn):
        """Validate the association generated"""
        super().validate(asn)

        if isinstance(asn, DMSBaseMixin):
            result = False
            try:
                result = all(test["validated"] for test in asn.validity.values())
            except (AttributeError, KeyError):
                raise AssociationNotValidError("Validation failed")
            if not result:
                raise AssociationNotValidError("Validation failed validity tests.")
        return True

    def _get_exposure(self):
        """Get string representation of the exposure id

        Returns
        -------
        exposure : str
            The Level3 Product name representation
            of the exposure & activity id.
        """
        exposure = ""
        try:
            activity_id = format_list(self.constraints["activity_id"].found_values)
        except KeyError:
            pass
        else:
            if activity_id not in _EMPTY:
                exposure = f"{activity_id:0>2s}"
        return exposure

    def _get_instrument(self):
        """Get string representation of the instrument

        Returns
        -------
        instrument : str
            The Level3 Product name representation
            of the instrument
        """
        instrument = format_list(self.constraints["instrument"].found_values)
        return instrument

    def _get_opt_element(self):
        """Get string representation of the optical elements

        Returns
        -------
        opt_elem : str
            The Level3 Product name representation
            of the optical elements.
        """
        # Retrieve all the optical elements
        opt_elems = []
        for opt_elem in ["opt_elem", "opt_elem2", "opt_elem3"]:
            try:
                values = list(self.constraints[opt_elem].found_values)
            except KeyError:
                pass
            else:
                values.sort(key=str.lower)
                value = format_list(values)
                if value not in _EMPTY:
                    opt_elems.append(value)

        # Build the string. Sort the elements in order to
        # create data-independent results
        opt_elems.sort(key=str.lower)
        opt_elem = "-".join(opt_elems)
        if opt_elem == "":
            opt_elem = "clear"

        return opt_elem

    def _get_subarray(self):
        """Get string representation of the subarray

        Returns
        -------
        subarray : str
            The Level3 Product name representation
            of the subarray.
        """
        result = ""
        try:
            subarray = format_list(self.constraints["seq_id"].found_values)
        except KeyError:
            subarray = None
            return result
        if subarray == 0:
            subarray = None
        try:
            format_list(self.constraints["subcat"].found_values)
        except KeyError:
            pass

        return result

    def _get_target(self):
        """Get string representation of the target

        Returns
        -------
        target : str
            The Level3 Product name representation
            of the target or source ID.
        """
        target_id = format_list(self.constraints["target"].found_values)
        target = f"t{str(target_id):0>3s}"
        return target

    def _get_grating(self):
        """Get string representation of the grating in use

        Returns
        -------
        grating : str
            The Level3 Product name representation
            of the grating in use.
        """
        grating_id = format_list(self.constraints["grating"].found_values)
        grating = f"{str(grating_id):0>3s}"
        return grating


# -----------------
# Basic constraints
# -----------------
class DMSAttrConstraint(AttrConstraint):
    """DMS-focused attribute constraint

    Forces definition of invalid values
    """

    def __init__(self, **kwargs):
        if kwargs.get("invalid_values", None) is None:
            kwargs["invalid_values"] = _EMPTY

        super().__init__(**kwargs)


class Constraint_TargetAcq(SimpleConstraint):
    """Select on target acquisition exposures

    Parameters
    ----------
    association:  ~romancal.associations.Association
        If specified, use the `get_exposure_type` method
        of the association rather than the utility version.
    """

    def __init__(self, association=None):
        if association is None:
            _get_exposure_type = get_exposure_type
        else:
            _get_exposure_type = association.get_exposure_type

        super().__init__(
            name="target_acq", value="target_acquisition", sources=_get_exposure_type
        )


class Constraint_WFSC(Constraint):
    """Match on Wave Front Sensing and Control Observations"""

    def __init__(self, *args, **kwargs):
        super().__init__(
            [
                Constraint(
                    [
                        DMSAttrConstraint(
                            name="wfsc",
                            sources=["visitype"],
                            value=".+wfsc.+",
                            force_unique=True,
                        )
                    ]
                )
            ]
        )


# #########
# Utilities
# #########
def format_list(alist):
    """Format a list according to DMS naming specs"""
    return "-".join(alist)


def get_exposure_type(item, default="science", association=None):
    """Determine the exposure type of a pool item

    Parameters
    ----------
    item : dict
        The pool entry to determine the exposure type of

    default : str or None
        The default exposure type.
        If None, routine will raise LookupError



    Returns
    -------
    exposure_type : str
        Exposure type. Can be one of

        - 'science': Item contains science data
        - 'target_acquisition': Item contains target acquisition data.
        - 'autoflat': NIRSpec AUTOFLAT
        - 'autowave': NIRSpec AUTOWAVE
        - 'psf': PSF
        - 'imprint': MSA/IFU Imprint/Leakcal

    Raises
    ------
    LookupError
        When `default` is None and an exposure type cannot be determined
    """

    # Specify how attributes of the item are retrieved.
    def _item_attr(item, sources):
        """Get attribute value of an item

        This simplifies the call to `item_getattr`
        """
        source, value = item_getattr(item, sources, association=association)
        return value

    # Define default type.
    result = default

    # Retrieve pointing type. This decides the basic exposure type.
    # If the pointing is not science, we're done.
    try:
        result = _item_attr(item, ["pntgtype"])
    except KeyError:
        pass
    else:
        if result != "science":
            return result

    # We have a science exposure. Refine further.
    #
    # Base type off of exposure type.
    try:
        exp_type = _item_attr(item, ["exp_type"])
    except KeyError:
        raise LookupError("Exposure type cannot be determined")

    result = EXPTYPE_MAP.get(exp_type, default)

    if result is None:
        raise LookupError("Cannot determine exposure type")

    # If result is not science, we're done.
    if result != "science":
        return result

    # For `science` data, compare against special modifiers
    # to further refine the type.
    for special, source in SPECIAL_EXPOSURE_MODIFIERS.items():
        try:
            _item_attr(item, source)
        except KeyError:
            pass
        else:
            result = special
            break

    return result


def item_getattr(item, attributes, association=None):
    """Return value from any of a list of attributes

    Parameters
    ----------
    item : dict
        item to retrieve from

    attributes : list
        List of attributes

    Returns
    -------
    (attribute, value)
        Returns the value and the attribute from
        which the value was taken.

    Raises
    ------
    KeyError
        None of the attributes are found in the dict.
    """
    if association is None:
        invalid_values = _EMPTY
    else:
        invalid_values = association.INVALID_VALUES
    return getattr_from_list(item, attributes, invalid_values=invalid_values)



import copy
import json
import os
import shutil
from io import StringIO
from pathlib import Path
from typing import Tuple

import numpy as np
import pytest
import requests
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from astropy.modeling.models import RotationSequence3D, Scale, Shift
from astropy.table import Table
from astropy.time import Time
from gwcs import coordinate_frames as cf
from gwcs import wcs
from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
from numpy.random import default_rng
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils
from stcal.tweakreg.astrometric_utils import get_catalog

from romancal.datamodels import ModelLibrary
from romancal.tweakreg import tweakreg_step as trs
from romancal.tweakreg.tweakreg_step import _validate_catalog_columns


class MockConnectionError:
    def __init__(self, *args, **kwargs):
        raise requests.exceptions.ConnectionError


@pytest.fixture()
def test_data_dir():
    return Path.joinpath(Path(__file__).parent, "data")


def create_custom_catalogs(tmp_path, base_image, catalog_format="ascii.ecsv"):
    """
    Creates a custom catalog with three datamodels and its respective catalog data.
    """
    img1 = base_image(shift_1=1000, shift_2=1000)
    img2 = base_image(shift_1=1010, shift_2=1010)
    img3 = base_image(shift_1=1020, shift_2=1020)

    img1.meta.filename = "img1.asdf"
    img2.meta.filename = "img2.asdf"
    img3.meta.filename = "img3.asdf"

    # create valid custom catalog data to be used with each input datamodel
    catalog_data1 = get_catalog_data(img1)
    catalog_data2 = get_catalog_data(img2)
    catalog_data3 = get_catalog_data(img3)

    custom_catalog_map = [
        {
            "cat_filename": "ref_catalog_1",
            "cat_datamodel": img1.meta.filename,
            "cat_data": catalog_data1,
        },
        {
            "cat_filename": "ref_catalog_2",
            "cat_datamodel": img2.meta.filename,
            "cat_data": catalog_data2,
        },
        {
            "cat_filename": "ref_catalog_3",
            "cat_datamodel": img3.meta.filename,
            "cat_data": catalog_data3,
        },
    ]

    # create catfile
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    for x in custom_catalog_map:
        # write line to catfile
        catfile_content.write(f"{x.get('cat_datamodel')} {x.get('cat_filename')}\n")
        # write out the catalog data
        t = Table(x.get("cat_data"), names=("x", "y"))
        t.write(tmp_path / x.get("cat_filename"), format=catalog_format)
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    return {"catfile": catfile, "datamodels": [img1, img2, img3]}


def create_asn_file(tmp_path, members_mapping=None):
    asn_content = """
        {
            "asn_type": "None",
            "asn_rule": "DMS_ELPP_Base",
            "version_id": null,
            "code_version": "0.9.1.dev28+ge987cc9.d20230106",
            "degraded_status": "No known degraded exposures in association.",
            "program": "noprogram",
            "constraints": "No constraints",
            "asn_id": "a3001",
            "target": "none",
            "asn_pool": "test_pool_name",
            "products": [
                {
                    "name": "files.asdf",
                    "members": [
                        {
                            "expname": "img_1.asdf",
                            "exptype": "science"
                        },
                        {
                            "expname": "img_2.asdf",
                            "exptype": "science"
                        }
                    ]
                }
            ]
        }
    """
    if members_mapping is not None:
        asn_dict = json.loads(asn_content)
        asn_dict["products"][0]["members"] = []
        for x in members_mapping:
            asn_dict["products"][0]["members"].append(x)
        asn_content = json.dumps(asn_dict)

    asn_file_path = str(tmp_path / "sample_asn.json")
    asn_file = StringIO()
    asn_file.write(asn_content)
    with open(asn_file_path, mode="w") as f:
        print(asn_file.getvalue(), file=f)

    return asn_file_path


def update_wcsinfo(input_dm):
    """
    Update WCSInfo with realistic data (i.e. obtained from romanisim simulations).

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    """
    input_dm.meta.wcsinfo.v2_ref = 0.42955128282521254
    input_dm.meta.wcsinfo.v3_ref = -0.2479976768255853
    input_dm.meta.wcsinfo.vparity = -1
    input_dm.meta.wcsinfo.v3yangle = -999999
    input_dm.meta.wcsinfo.ra_ref = 270.0
    input_dm.meta.wcsinfo.dec_ref = 66.0
    input_dm.meta.wcsinfo.roll_ref = 60
    input_dm.meta.wcsinfo.s_region = (
        "POLYGON ICRS "
        "269.3318903230621 65.56866666048172 "
        "269.32578768154605 65.69246311613287 "
        "269.02457173246125 65.69201346248587 "
        "269.0333096074621 65.56870823657276 "
    )


def _create_tel2sky_model(input_dm):
    """
    Create the transform from telescope to sky.

    The transform is defined with a reference point in a Frame
    associated with the telescope (V2, V3) in arcsec, the corresponding
    reference point on sky (RA_REF, DEC_REF) in deg, and the position angle
    at the center of the aperture, ROLL_REF in deg.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.

    Returns
    -------
    astropy.modeling.core.CompoundModel
        An astropy compound model with the transformations to map
        the telescope reference system onto the world reference systems.
    """
    v2_ref = input_dm.meta.wcsinfo.v2_ref / 3600
    v3_ref = input_dm.meta.wcsinfo.v3_ref / 3600
    roll_ref = input_dm.meta.wcsinfo.roll_ref
    ra_ref = input_dm.meta.wcsinfo.ra_ref
    dec_ref = input_dm.meta.wcsinfo.dec_ref

    angles = np.array([v2_ref, -v3_ref, roll_ref, dec_ref, -ra_ref])
    axes = "zyxyz"
    rot = RotationSequence3D(angles, axes_order=axes)

    # The sky rotation expects values in deg.
    # This should be removed when models work with quantities.
    model = (
        (Scale(0.1 / 3600) & Scale(0.1 / 3600))
        | SphericalToCartesian(wrap_lon_at=180)
        | rot
        | CartesianToSpherical(wrap_lon_at=360)
    )
    model.name = "v23tosky"
    return model


def create_basic_wcs(
    img_shape: tuple = (100, 100),
    ref_pix: tuple = (0, 0),
    ref_val: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("10 deg"),
        u.Quantity("0 deg"),
    ),
    pix_scale: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("0.1 arcsec"),
        u.Quantity("0.1 arcsec"),
    ),
    theta: u.Quantity = u.Quantity("0 deg"),
):
    """
    Creates a basic WCS (no distortion) to map pixel coordinates
    onto celestial spherical coordinates.

    Parameters
    ----------
    img_shape : tuple, optional
        Dimension of the pixel coordinates, by default (100, 100).
    ref_pix : tuple, optional
        Projection plane coordinates of the reference pixel, by default (0, 0)
    ref_val : tuple[u.Quantity, u.Quantity], optional
        Celestial coordinates at the reference pixel in degrees,
        by default (10, 0)*u.degree
    pix_scale : tuple[u.Quantity, u.Quantity], optional
        Size of a pixel side in arcsec, by default (0.1, 0.1)*u.arcsec
    theta : u.Quantity, optional
        CCW rotation angle of the projection plane, by default 0*u.deg

    Returns
    -------
    type: gwcs.wcs.WCS object
        A gwcs.wcs.WCS object with a model f : (x, y) |-> (ra, dec).

    Notes
    -----
    With the default parameters, the result will be a WCS for an image
    that's 100x100 pixels (@0.1"/pix in both axis) and reference
    coordinates (ra=10 deg, dec=0 deg, ICRS) sitting at the origin of the
    pixel coordinates frame (0,0).

    The WCS object created by this method generates a pipeline that
    does not contain the required steps to validate against the
    TweakReg pipeline.
    """

    # linear transformations
    shift_pixel_coords = models.Shift(-ref_pix[0]) & models.Shift(-ref_pix[1])
    scaling_matrix = np.array(
        [[pix_scale.to("deg")[0].value, 0], [0, pix_scale.to("deg")[1].value]]
    )
    rot_matrix = np.array(
        [
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)],
        ]
    )
    matrix = np.dot(rot_matrix, scaling_matrix)
    affine_transf = models.AffineTransformation2D(matrix)
    affine_transf.inverse = models.AffineTransformation2D(np.linalg.inv(matrix))
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        ref_val[0].value, ref_val[1].value, 180
    )
    det2sky = shift_pixel_coords | affine_transf | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector", unit=(u.pix, u.pix), axes_names=("x", "y")
    )
    cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    sky_frame = cf.CelestialFrame(
        name="sky", unit=(u.deg, u.deg), reference_frame=coord.ICRS()
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    wcsobj = wcs.WCS(pipeline)
    wcsobj.bounding_box = (
        (-0.5, img_shape[0] + 0.5),
        (-0.5, img_shape[1] + 0.5),
    )

    return wcsobj


def create_wcs_for_tweakreg_pipeline(input_dm, shift_1=0, shift_2=0):
    """
    Create a basic WCS object (with optional shift) and
    append it to the input_dm.meta attribute.

    The WCS object will have a pipeline with the required
    steps to validate against the TweakReg pipeline.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    shift_1 : int, optional
        Shift to be applied in the direction of the first axis, by default 0
    shift_2 : int, optional
        Shift to be applied in the direction of the second axis, by default 0
    """

    shape = input_dm.data.shape

    # create necessary transformations
    distortion = Shift(-shift_1) & Shift(-shift_2)
    distortion.bounding_box = ((-0.5, shape[-1] + 0.5), (-0.5, shape[-2] + 0.5))
    tel2sky = _create_tel2sky_model(input_dm)

    # create required frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # create pipeline
    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, tel2sky),
        wcs.Step(world, None),
    ]

    wcs_obj = wcs.WCS(pipeline)

    input_dm.meta["wcs"] = wcs_obj


def get_catalog_data(input_dm, **kwargs):
    ra = kwargs.get("ra", 270)
    dec = kwargs.get("dec", 66)
    sr = kwargs.get("sr", 100 / 3600)
    add_shifts = kwargs.get("add_shifts", False)
    gaia_cat = get_catalog(right_ascension=ra, declination=dec, search_radius=sr)
    gaia_source_coords = [(ra, dec) for ra, dec in zip(gaia_cat["ra"], gaia_cat["dec"])]
    catalog_data = np.array(
        [input_dm.meta.wcs.world_to_pixel(ra, dec) for ra, dec in gaia_source_coords]
    )
    if add_shifts:
        rng = np.random.default_rng(seed=int(ra + dec))
        shifts = rng.uniform(-1, 1, size=catalog_data.shape)
        catalog_data += shifts
    return catalog_data


def create_base_image_source_catalog(
    tmp_path,
    output_filename,
    catalog_data,
    catalog_format: str = "ascii.ecsv",
    save_catalogs=True,
):
    """
    Write a temp CSV file to be used as source catalog, similar to what
    is produced by the previous pipeline step, source detection.

    Parameters
    ----------
    tmp_path : pathlib.PosixPath
        A path-like object representing the path where to save the file.
    output_filename : string
        The output filename (with extension).
    catalog_data : numpy.ndarray
        A numpy array with the (x, y) coordinates of the
        "detected" sources
    catalog_format : str, optional
        A string indicating the catalog format.
    save_catalogs : boolean, optional
        A boolean indicating whether the source catalog should be saved to disk.
    """
    src_detector_coords = catalog_data
    output = os.path.join(tmp_path, output_filename)
    t = Table(src_detector_coords, names=("x", "y"))
    if save_catalogs:
        t.write((tmp_path / output), format=catalog_format)
    # mimic the same output format from SourceDetectionStep
    t.add_column([i for i in range(len(t))], name="id", index=0)
    t.add_column([np.float64(i) for i in range(len(t))], name="flux")
    t.rename_columns(["x", "y"], ["xcentroid", "ycentroid"])
    return t.as_array()


def add_tweakreg_catalog_attribute(
    tmp_path,
    input_dm,
    catalog_filename="base_image_sources",
    catalog_data=None,
    catalog_format: str = "ascii.ecsv",
    save_catalogs=True,
    **kwargs,
):
    """
    Add tweakreg_catalog attribute to the meta, which is a mandatory
    attribute for TweakReg. The tweakreg_catalog attribute contains
    a table with the coordinates of the sources detected by the
    previous step. The coordinates will be used in the process of
    correcting the original WCS.

    Parameters
    ----------
    tmp_path : pathlib.PosixPath
        A path-like object representing the path where to save the file.
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    catalog_filename : str
        Filename to be used when saving the catalog to disk.
    catalog_data : numpy.ndarray, optional
        A numpy array with the (x, y) coordinates of the
        "detected" sources, by default None (see note below).
    catalog_format : str, optional
        A string indicating the catalog format.
    save_catalogs : boolean, optional
        A boolean indicating whether the source catalog should be saved to disk.

    Notes
    ----
    - kwargs will be passed on to get_catalog_data();
    - if no catalog_data is provided, a default catalog will be created
    by fetching data from Gaia within a search radius of 100 arcsec
    centered at RA=270, Dec=66.
    """
    tweakreg_catalog_filename = catalog_filename
    if catalog_data is None:
        catalog_data = get_catalog_data(input_dm, **kwargs)

    source_catalog = create_base_image_source_catalog(
        tmp_path,
        tweakreg_catalog_filename,
        catalog_data=catalog_data,
        catalog_format=catalog_format,
        save_catalogs=save_catalogs,
    )

    input_dm.meta["source_detection"] = maker_utils.mk_source_detection()

    if save_catalogs:
        # SourceDetectionStep adds the catalog path+filename
        input_dm.meta.source_detection["tweakreg_catalog_name"] = os.path.join(
            tmp_path, tweakreg_catalog_filename
        )
    else:
        # SourceDetectionStep attaches the catalog data as a structured array
        input_dm.meta.source_detection["tweakreg_catalog"] = source_catalog


@pytest.fixture
def base_image():
    """
    Create a base image with a realistic WCSInfo and a WCS.

    Notes
    -----
    The size of the image needs to be relatively large in order for
    the source catalog step to find a reasonable number of sources in the image.

    shift_1 and shift_2 (units in pixel) are used to shift the WCS projection plane.
    """

    def _base_image(shift_1=0, shift_2=0):
        l2 = maker_utils.mk_level2_image(shape=(2000, 2000))
        l2.meta.exposure.mid_time = Time("2016-01-01T00:00:00")
        # update wcsinfo
        update_wcsinfo(l2)
        # add a dummy WCS object
        create_wcs_for_tweakreg_pipeline(l2, shift_1=shift_1, shift_2=shift_2)
        l2_im = rdm.ImageModel(l2)
        return l2_im

    return _base_image


@pytest.mark.parametrize(
    "input, error_type",
    [
        (list(), (Exception,)),
        ([""], (Exception,)),
        (["", ""], (Exception,)),
        ("", (Exception,)),
        ([1, 2, 3], (Exception,)),
    ],
)
def test_tweakreg_raises_error_on_invalid_input(input, error_type):
    # sourcery skip: list-literal
    """Test that TweakReg raises an error when an invalid input is provided."""
    with pytest.raises(error_type):
        trs.TweakRegStep.call(input)


def test_tweakreg_raises_attributeerror_on_missing_tweakreg_catalog(base_image):
    """
    Test that TweakReg raises an AttributeError if meta.tweakreg_catalog is missing.
    """
    img = base_image()
    # remove attribute
    del img.meta.source_detection["tweakreg_catalog_name"]
    with pytest.raises(AttributeError):
        trs.TweakRegStep.call([img])


def test_tweakreg_returns_modellibrary_on_roman_datamodel_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelLibrary when processing an open Roman DataModel as input."""

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    test_input = img

    res = trs.TweakRegStep.call(test_input)
    assert isinstance(res, ModelLibrary)
    with res:
        model = res.borrow(0)
        assert model.meta.cal_step.tweakreg == "COMPLETE"
        res.shelve(model, 0, modify=False)


def test_tweakreg_returns_modellibrary_on_modellibrary_as_input(tmp_path, base_image):
    """Test that TweakReg always returns a ModelLibrary when processing a ModelLibrary as input."""

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    test_input = ModelLibrary([img])

    res = trs.TweakRegStep.call(test_input)
    assert isinstance(res, ModelLibrary)
    with res:
        model = res.borrow(0)
        assert model.meta.cal_step.tweakreg == "COMPLETE"
        res.shelve(model, 0, modify=False)


def test_tweakreg_returns_modellibrary_on_association_file_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelLibrary when processing an association file as input."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")
    asn_filepath = create_asn_file(tmp_path)

    test_input = asn_filepath

    res = trs.TweakRegStep.call(test_input)
    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert model.meta.cal_step.tweakreg == "COMPLETE"
            res.shelve(model, i, modify=False)


def test_tweakreg_returns_modellibrary_on_list_of_asdf_file_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelLibrary when processing a list of ASDF files as input."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")

    tmp_path_str = tmp_path.as_posix()
    test_input = [
        f"{tmp_path_str}/img_1.asdf",
        f"{tmp_path_str}/img_2.asdf",
    ]

    res = trs.TweakRegStep.call(test_input)
    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert model.meta.cal_step.tweakreg == "COMPLETE"
            res.shelve(model, i, modify=False)


def test_tweakreg_returns_modellibrary_on_list_of_roman_datamodels_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelLibrary when processing a list of open Roman datamodels as input."""
    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")

    test_input = [img_1, img_2]

    res = trs.TweakRegStep.call(test_input)
    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert model.meta.cal_step.tweakreg == "COMPLETE"
            res.shelve(model, i, modify=False)


def test_tweakreg_updates_cal_step(tmp_path, base_image):
    """Test that TweakReg updates meta.cal_step with tweakreg = COMPLETE."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)
    res = trs.TweakRegStep.call([img])

    with res:
        model = res.borrow(0)
        assert hasattr(model.meta.cal_step, "tweakreg")
        assert model.meta.cal_step.tweakreg == "COMPLETE"
        res.shelve(model, 0, modify=False)


def test_tweakreg_updates_group_id(tmp_path, base_image):
    """Test that TweakReg updates 'group_id' with a non-zero length string."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)
    res = trs.TweakRegStep.call([img])

    with res:
        model = res.borrow(0)
        assert hasattr(model.meta, "group_id")
        res.shelve(model, 0, modify=False)


@pytest.mark.parametrize(
    "abs_refcat",
    (
        "GAIADR1",
        "GAIADR2",
        "GAIADR3",
    ),
)
def test_tweakreg_save_valid_abs_refcat(tmp_path, abs_refcat, request):
    """Test that TweakReg saves the catalog used for absolute astrometry."""

    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    catalog_filename = "ref_catalog.ecsv"
    abs_refcat_filename = f"fit_{abs_refcat.lower()}_ref.ecsv"
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename=catalog_filename)

    trs.TweakRegStep.call(
        [img],
        save_abs_catalog=True,
        abs_refcat=abs_refcat,
        catalog_path=str(tmp_path),
        output_dir=str(tmp_path),
    )

    assert os.path.exists(tmp_path / abs_refcat_filename)


@pytest.mark.parametrize(
    "abs_refcat",
    (None, ""),
)
def test_tweakreg_defaults_to_valid_abs_refcat(tmp_path, abs_refcat, request):
    """Test that TweakReg defaults to DEFAULT_ABS_REFCAT on invalid values."""

    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    catalog_filename = "ref_catalog.ecsv"
    abs_refcat_filename = f"fit_{trs.DEFAULT_ABS_REFCAT.lower()}_ref.ecsv"
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename=catalog_filename)

    trs.TweakRegStep.call(
        [img],
        save_abs_catalog=True,
        abs_refcat=abs_refcat,
        catalog_path=str(tmp_path),
        output_dir=str(tmp_path),
    )

    assert os.path.exists(tmp_path / abs_refcat_filename)


def test_tweakreg_raises_error_on_invalid_abs_refcat(tmp_path, base_image):
    """Test that TweakReg raises an error when an invalid abs_refcat is provided."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    with pytest.raises(TypeError):
        trs.TweakRegStep.call(
            [img],
            save_abs_catalog=True,
            abs_refcat="my_ref_cat",
            catalog_path=str(tmp_path),
            output_dir=str(tmp_path),
        )


def test_tweakreg_combine_custom_catalogs_and_asn_file(tmp_path, base_image):
    """
    Test that TweakRegStep can handle a custom catalog for the members of an ASN file.
    In this case, the user can create a custom catalog file (catfile) for each of the
    members of an ASN file.
    """
    # create custom catalog and input datamodels
    catalog_format = "ascii.ecsv"
    res_dict = create_custom_catalogs(
        tmp_path, base_image, catalog_format=catalog_format
    )
    catfile = res_dict.get("catfile")
    img1, img2, img3 = res_dict.get("datamodels")
    add_tweakreg_catalog_attribute(tmp_path, img1, catalog_filename="img1")
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")
    add_tweakreg_catalog_attribute(tmp_path, img3, catalog_filename="img3")
    img1.save(tmp_path / "img1.asdf")
    img2.save(tmp_path / "img2.asdf")
    img3.save(tmp_path / "img3.asdf")

    # create ASN file
    asn_filepath = create_asn_file(
        tmp_path,
        members_mapping=[
            {"expname": img1.meta.filename, "exptype": "science"},
            {"expname": img2.meta.filename, "exptype": "science"},
            {"expname": img3.meta.filename, "exptype": "science"},
        ],
    )
    with open(asn_filepath) as f:
        asn_content = json.load(f)

    res = trs.TweakRegStep.call(
        asn_filepath,
        use_custom_catalogs=True,
        catalog_format=catalog_format,
        catfile=catfile,
    )

    assert isinstance(res, ModelLibrary)

    with res:
        for i, (model, target) in enumerate(zip(res, [img1, img2, img3])):
            assert hasattr(model.meta, "asn")

            assert (
                model.meta["exptype"]
                == asn_content["products"][0]["members"][i]["exptype"]
            )

            assert model.meta.filename == target.meta.filename

            assert type(model) is type(target)

            assert (model.data == target.data).all()

            res.shelve(model, i, modify=False)


@pytest.mark.parametrize(
    "theta, offset_x, offset_y",
    [
        (0 * u.deg, 0, 0),
        (0 * u.deg, 1, 0),
        (0 * u.deg, 0, 1),
        (0 * u.deg, 1, 1),
        (0.05 * u.deg, 0, 0),
        (0.05 * u.deg, 1, 0),
        (0.05 * u.deg, 0, 1),
        (0.05 * u.deg, 1, 1),
        (0.1 * u.deg, 0, 0),
        (0.1 * u.deg, 1, 0),
        (0.1 * u.deg, 0, 1),
        (0.1 * u.deg, 1, 1),
    ],
)
def test_tweakreg_rotated_plane(tmp_path, theta, offset_x, offset_y, request):
    """
    Test that TweakReg returns accurate results.
    """
    gaia_cat = get_catalog(
        right_ascension=270, declination=66, search_radius=100 / 3600
    )
    gaia_source_coords = [(ra, dec) for ra, dec in zip(gaia_cat["ra"], gaia_cat["dec"])]

    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    original_wcs = copy.deepcopy(img.meta.wcs)

    # calculate original (x,y) for Gaia sources
    original_xy_gaia_sources = np.array(
        [original_wcs.world_to_pixel(ra, dec) for ra, dec in gaia_source_coords]
    )
    # move Gaia sources around by applying linear transformations
    # to their coords in the projected plane (same as a "wrong WCS")
    rot_matrix = models.Rotation2D(angle=theta)
    transformed_xy_gaia_sources = np.array(
        [
            rot_matrix(x, y) + np.array([offset_x, offset_y])
            for x, y in original_xy_gaia_sources
        ]
    )
    # save modified catalog to meta.tweakreg_catalog
    # (coords in the projection plane)
    add_tweakreg_catalog_attribute(
        tmp_path, img, catalog_data=transformed_xy_gaia_sources
    )

    trs.TweakRegStep.call([img], abs_minobj=3)

    # get world coords for Gaia sources using "wrong WCS"
    original_ref_source = [
        original_wcs.pixel_to_world(x, y) for x, y in transformed_xy_gaia_sources
    ]
    # get world coords for Gaia sources using tweaked WCS
    new_ref_source = [
        img.meta.wcs.pixel_to_world(x, y) for x, y in transformed_xy_gaia_sources
    ]
    # celestial coordinates for Gaia sources
    gaia_ref_source = [
        coord.SkyCoord(ra * u.deg, dec * u.deg, frame="icrs")
        for ra, dec in gaia_source_coords
    ]
    # calculate distance between "wrong WCS" result and Gaia
    # (rounded to the 10th decimal place to avoid floating point issues)
    dist1 = [
        np.round(gref.separation(oref), 10)
        for gref, oref in zip(gaia_ref_source, original_ref_source)
    ]
    # calculate distance between tweaked WCS result and Gaia
    # (rounded to the 10th decimal place to avoid floating point issues)
    dist2 = [
        np.round(gref.separation(nref), 10)
        for gref, nref in zip(gaia_ref_source, new_ref_source)
    ]

    assert np.array([np.less_equal(d2, d1) for d1, d2 in zip(dist1, dist2)]).all()


def test_tweakreg_parses_asn_correctly(tmp_path, base_image):
    """Test that TweakReg can parse an ASN file properly."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    img_1.meta["filename"] = "img_1.asdf"
    img_2.meta["filename"] = "img_2.asdf"
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")
    asn_filepath = create_asn_file(tmp_path)
    with open(asn_filepath) as f:
        asn_content = json.load(f)

    res = trs.TweakRegStep.call(asn_filepath)

    assert isinstance(res, ModelLibrary)

    with res:
        models = list(res)
        assert hasattr(models[0].meta, "asn")
        assert (
            models[0].meta["exptype"]
            == asn_content["products"][0]["members"][0]["exptype"]
        )
        assert (
            models[1].meta["exptype"]
            == asn_content["products"][0]["members"][1]["exptype"]
        )
        assert models[0].meta.asn["pool_name"] == asn_content["asn_pool"]
        assert models[1].meta.asn["pool_name"] == asn_content["asn_pool"]

        assert models[0].meta.filename == img_1.meta.filename
        assert models[1].meta.filename == img_2.meta.filename

        assert type(models[0]) is type(img_1)
        assert type(models[1]) is type(img_2)

        assert (models[0].data == img_1.data).all()
        assert (models[1].data == img_2.data).all()

        [res.shelve(m, i, modify=False) for i, m in enumerate(models)]


def test_fit_results_in_meta(tmp_path, base_image):
    """
    Test that the WCS fit results from tweakwcs are available in the meta tree.
    """

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    res = trs.TweakRegStep.call([img])

    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert hasattr(model.meta, "wcs_fit_results")
            assert len(model.meta.wcs_fit_results) > 0
            res.shelve(model, i, modify=False)


def test_tweakreg_handles_multiple_groups(tmp_path, base_image):
    """
    Test that TweakRegStep can perform relative alignment for all images in the groups
    before performing absolute alignment.
    """
    img1 = base_image(shift_1=1000, shift_2=1000)
    img2 = base_image(shift_1=990, shift_2=990)
    add_tweakreg_catalog_attribute(tmp_path, img1, catalog_filename="img1")
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")

    img1.meta.observation.program = 1
    img1.meta.observation["observation_id"] = "1"
    img2.meta.observation.program = 2
    img2.meta.observation["observation_id"] = "2"

    img1.meta["filename"] = "file1.asdf"
    img2.meta["filename"] = "file2.asdf"

    res = trs.TweakRegStep.call([img1, img2])

    assert len(res.group_names) == 2
    with res:
        for r, i in zip(res, [img1, img2]):
            assert r.meta.group_id == i.meta.observation.observation_id
            res.shelve(r, modify=False)


def test_parse_catfile_valid_catalog(tmp_path, base_image):
    """
    Test that _parse_catfile can parse a custom catalog with valid format.
    """
    # create custom catalog file and input datamodels
    catalog_format = "ascii.ecsv"
    res_dict = create_custom_catalogs(
        tmp_path, base_image, catalog_format=catalog_format
    )
    catfile = res_dict.get("catfile")
    catdict = trs._parse_catfile(catfile)

    assert all(
        x.meta.filename == y for x, y in zip(res_dict.get("datamodels"), catdict.keys())
    )


@pytest.mark.parametrize("catfile", (None, ""))
def test_parse_catfile_returns_none(catfile):
    """
    Test that _parse_catfile returns None when catfile = None or catfile = "".
    """
    catdict = trs._parse_catfile(catfile=catfile)

    assert catdict is None


@pytest.mark.parametrize(
    "catfile_line_content",
    ["img1.asdf\nimg2.asdf\nimg3.asdf"],
)
def test_parse_catfile_returns_none_on_invalid_content(tmp_path, catfile_line_content):
    """
    Test that _parse_catfile returns a dict where all the values are None
    if only filename is present in catfile (i.e. no associated catalog).
    """
    # create custom catalog file and input datamodels
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    # write empty line to catfile
    catfile_content.write(catfile_line_content)
    # write StringIO object to disk
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    catdict = trs._parse_catfile(catfile)

    assert not all(catdict.values())


@pytest.mark.parametrize(
    "catfile_line_content",
    ["img1.asdf column1 column2 column3"],
)
def test_parse_catfile_raises_error_on_invalid_content(tmp_path, catfile_line_content):
    """
    Test that _parse_catfile raises an error if catfile contains more than
    two columns (i.e. image name and corresponding catalog path).
    """
    # create custom catalog file and input datamodels
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    # write empty line to catfile
    catfile_content.write(catfile_line_content)
    # write StringIO object to disk
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    with pytest.raises(ValueError):
        trs._parse_catfile(catfile)


def test_update_source_catalog_coordinates(tmp_path, base_image):
    """Test that TweakReg updates the catalog coordinates with the tweaked WCS."""

    os.chdir(tmp_path)

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    tweakreg = trs.TweakRegStep()

    # create SourceCatalogModel
    source_catalog_model = setup_source_catalog_model(img)

    # save SourceCatalogModel
    tweakreg.save_model(
        source_catalog_model,
        output_file="img_1.asdf",
        suffix="cat",
        force=True,
    )

    # update tweakreg catalog name
    img.meta.source_detection.tweakreg_catalog_name = "img_1_cat.asdf"

    # run TweakRegStep
    res = trs.TweakRegStep.call([img])

    # tweak the current WCS using TweakRegStep and save the updated cat file
    with res:
        dm = res.borrow(0)
        assert dm.meta.source_detection.tweakreg_catalog_name == "img_1_cat.asdf"
        tweakreg.update_catalog_coordinates(
            dm.meta.source_detection.tweakreg_catalog_name, dm.meta.wcs
        )
        res.shelve(dm, 0)

    # read in saved catalog coords
    cat = rdm.open("img_1_cat.asdf")
    cat_ra_centroid = cat.source_catalog["ra_centroid"]
    cat_dec_centroid = cat.source_catalog["dec_centroid"]
    cat_ra_psf = cat.source_catalog["ra_psf"]
    cat_dec_psf = cat.source_catalog["dec_psf"]

    # calculate world coords using tweaked WCS
    expected_centroid = img.meta.wcs(
        cat.source_catalog["xcentroid"], cat.source_catalog["ycentroid"]
    )
    expected_psf = img.meta.wcs(
        cat.source_catalog["x_psf"], cat.source_catalog["y_psf"]
    )

    # compare coordinates (make sure tweaked WCS was applied to cat file coords)
    np.testing.assert_array_equal(cat_ra_centroid, expected_centroid[0])
    np.testing.assert_array_equal(cat_dec_centroid, expected_centroid[1])
    np.testing.assert_array_equal(cat_ra_psf, expected_psf[0])
    np.testing.assert_array_equal(cat_dec_psf, expected_psf[1])


def test_source_catalog_coordinates_have_changed(tmp_path, base_image):
    """Test that the original catalog file content is different from the updated file."""

    os.chdir(tmp_path)

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    tweakreg = trs.TweakRegStep()

    # create SourceCatalogModel
    source_catalog_model = setup_source_catalog_model(img)

    # save SourceCatalogModel
    tweakreg.save_model(
        source_catalog_model,
        output_file="img_1.asdf",
        suffix="cat",
        force=True,
    )
    # save original data
    shutil.copy("img_1_cat.asdf", "img_1_cat_original.asdf")

    # update tweakreg catalog name
    img.meta.source_detection.tweakreg_catalog_name = "img_1_cat.asdf"

    # run TweakRegStep
    res = trs.TweakRegStep.call([img])

    # tweak the current WCS using TweakRegStep and save the updated cat file
    with res:
        dm = res.borrow(0)
        assert dm.meta.source_detection.tweakreg_catalog_name == "img_1_cat.asdf"
        tweakreg.update_catalog_coordinates(
            dm.meta.source_detection.tweakreg_catalog_name, dm.meta.wcs
        )
        res.shelve(dm, 0)

    cat_original = rdm.open("img_1_cat_original.asdf")
    cat_updated = rdm.open("img_1_cat.asdf")

    # set max absolute and relative tolerance to ~ 1/2 a pixel
    atol = u.Quantity(0.11 / 2, "arcsec").to("deg").value
    rtol = 5e-8

    # testing that nothing moved by more than 1/2 a pixel
    assert np.allclose(
        cat_original.source_catalog["ra_centroid"],
        cat_updated.source_catalog["ra_centroid"],
        atol=atol,
        rtol=rtol,
    )
    assert np.allclose(
        cat_original.source_catalog["dec_centroid"],
        cat_updated.source_catalog["dec_centroid"],
        atol=atol,
        rtol=rtol,
    )
    assert np.allclose(
        cat_original.source_catalog["ra_psf"],
        cat_updated.source_catalog["ra_psf"],
        atol=atol,
        rtol=rtol,
    )
    assert np.allclose(
        cat_original.source_catalog["dec_psf"],
        cat_updated.source_catalog["dec_psf"],
        atol=atol,
        rtol=rtol,
    )
    # testing that things did move by more than ~ 1/100 of a pixel
    assert not np.allclose(
        cat_original.source_catalog["ra_centroid"],
        cat_updated.source_catalog["ra_centroid"],
        atol=atol / 100,
        rtol=rtol / 100,
    )
    assert not np.allclose(
        cat_original.source_catalog["dec_centroid"],
        cat_updated.source_catalog["dec_centroid"],
        atol=atol / 100,
        rtol=rtol / 100,
    )
    assert not np.allclose(
        cat_original.source_catalog["ra_psf"],
        cat_updated.source_catalog["ra_psf"],
        atol=atol / 100,
        rtol=rtol / 100,
    )
    assert not np.allclose(
        cat_original.source_catalog["dec_psf"],
        cat_updated.source_catalog["dec_psf"],
        atol=atol / 100,
        rtol=rtol / 100,
    )


def setup_source_catalog_model(img):
    """
    Set up the source catalog model.

    Notes
    -----
    This function reads the source catalog from a file, renames columns to match
    expected names, adds mock PSF coordinates, applies random shifts to the centroid
    and PSF coordinates, and calculates the world coordinates for the centroids.
    """
    cat_model = rdm.SourceCatalogModel
    source_catalog_model = maker_utils.mk_datamodel(cat_model)
    # this will be the output filename
    source_catalog_model.meta.filename = "img_1.asdf"

    # read in the mock table
    source_catalog = Table.read("img_1", format="ascii.ecsv")
    # rename columns to match expected column names
    source_catalog.rename_columns(["x", "y"], ["xcentroid", "ycentroid"])
    # add mock PSF coordinates
    source_catalog["x_psf"] = source_catalog["xcentroid"]
    source_catalog["y_psf"] = source_catalog["ycentroid"]

    # generate a set of random shifts to be added to the original coordinates
    seed = 13
    rng = default_rng(seed)
    shift_x = rng.uniform(-0.5, 0.5, size=len(source_catalog))
    shift_y = rng.uniform(-0.5, 0.5, size=len(source_catalog))
    # add random fraction of a pixel shifts to the centroid coordinates
    source_catalog["xcentroid"] += shift_x
    source_catalog["ycentroid"] += shift_y

    # generate another set of random shifts to be added to the original coordinates
    seed = 5
    rng = default_rng(seed)
    shift_x = rng.uniform(-0.5, 0.5, size=len(source_catalog))
    shift_y = rng.uniform(-0.5, 0.5, size=len(source_catalog))
    # add random fraction of a pixel shifts to the centroid coordinates
    source_catalog["x_psf"] += shift_x
    source_catalog["y_psf"] += shift_y

    # calculate centroid world coordinates
    centroid = img.meta.wcs(
        source_catalog["xcentroid"],
        source_catalog["ycentroid"],
    )
    # calculate PSF world coordinates
    psf = img.meta.wcs(
        source_catalog["x_psf"],
        source_catalog["y_psf"],
    )
    # add world coordinates to catalog
    source_catalog["ra_centroid"], source_catalog["dec_centroid"] = centroid
    source_catalog["ra_psf"], source_catalog["dec_psf"] = psf
    # add units
    source_catalog["ra_centroid"].unit = u.deg
    source_catalog["dec_centroid"].unit = u.deg
    source_catalog["ra_psf"].unit = u.deg
    source_catalog["dec_psf"].unit = u.deg

    # add source catalog to SourceCatalogModel
    source_catalog_model.source_catalog = source_catalog

    return source_catalog_model


@pytest.mark.parametrize(
    "exposure_type",
    ["WFI_GRISM", "WFI_PRISM", "WFI_DARK", "WFI_FLAT", "WFI_WFSC"],
)
def test_tweakreg_skips_invalid_exposure_types(exposure_type, tmp_path, base_image):
    """Test that TweakReg updates meta.cal_step with tweakreg = COMPLETE."""
    img1 = base_image(shift_1=1000, shift_2=1000)
    img1.meta.exposure.type = exposure_type
    img2 = base_image(shift_1=1000, shift_2=1000)
    img2.meta.exposure.type = exposure_type
    res = trs.TweakRegStep.call([img1, img2])

    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert hasattr(model.meta.cal_step, "tweakreg")
            assert model.meta.cal_step.tweakreg == "SKIPPED"
            res.shelve(model, i, modify=False)


@pytest.mark.parametrize(
    "catalog_data, expected_colnames, raises_exception",
    [
        # both 'x' and 'y' columns present
        ({"x": [1, 2, 3], "y": [4, 5, 6]}, ["x", "y"], False),
        # 'xcentroid' and 'ycentroid' columns present, should be renamed
        ({"xcentroid": [1, 2, 3], "ycentroid": [4, 5, 6]}, ["x", "y"], False),
        # 'x' present, 'ycentroid' present, should rename 'ycentroid' to 'y'
        ({"x": [1, 2, 3], "ycentroid": [4, 5, 6]}, ["x", "y"], False),
        # 'xcentroid' present, 'y' present, should rename 'xcentroid' to 'x'
        ({"xcentroid": [1, 2, 3], "y": [4, 5, 6]}, ["x", "y"], False),
        # neither 'x' nor 'xcentroid' present
        ({"y": [4, 5, 6]}, None, True),
        # neither 'y' nor 'ycentroid' present
        ({"x": [1, 2, 3]}, None, True),
        # no relevant columns present
        (
            {"a": [1, 2, 3], "b": [4, 5, 6]},
            None,
            True,
        ),
    ],
)
def test_validate_catalog_columns(catalog_data, expected_colnames, raises_exception):
    """Test that TweakRegStep._validate_catalog_columns() correctly validates the
    presence of required columns ('x' and 'y') in the provided catalog."""
    catalog = Table(catalog_data)
    if raises_exception:
        with pytest.raises(ValueError):
            _validate_catalog_columns(catalog)
    else:
        _validate_catalog_columns(catalog)
        assert set(catalog.colnames) == set(expected_colnames)


def test_tweakreg_handles_mixed_exposure_types(tmp_path, base_image):
    """Test that TweakReg can handle mixed exposure types
    (non-WFI_IMAGE data will be marked as SKIPPED only and won't be processed)."""
    img1 = base_image(shift_1=1000, shift_2=1000)
    img1.meta.exposure.type = "WFI_GRISM"

    img2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")
    img2.meta.exposure.type = "WFI_IMAGE"

    img3 = base_image(shift_1=1000, shift_2=1000)
    img3.meta.exposure.type = "WFI_GRISM"

    img4 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img4, catalog_filename="img4")
    img4.meta.exposure.type = "WFI_IMAGE"

    img5 = base_image(shift_1=1000, shift_2=1000)
    img5.meta.exposure.type = "WFI_GRISM"

    res = trs.TweakRegStep.call([img1, img2, img3, img4, img5])

    assert len(res) == 5
    assert img1.meta.cal_step.tweakreg == "SKIPPED"
    assert img2.meta.cal_step.tweakreg == "COMPLETE"
    assert img3.meta.cal_step.tweakreg == "SKIPPED"
    assert img4.meta.cal_step.tweakreg == "COMPLETE"
    assert img5.meta.cal_step.tweakreg == "SKIPPED"


def test_tweakreg_updates_s_region(tmp_path, base_image):
    """Test that the TweakRegStep updates the s_region attribute."""
    img = base_image(shift_1=1000, shift_2=1000)
    old_fake_s_region = "POLYGON ICRS 1.0000000000000 2.0000000000000 3.0000000000000 4.0000000000000 5.0000000000000 6.0000000000000 7.0000000000000 8.0000000000000 "
    img.meta.wcsinfo["s_region"] = old_fake_s_region
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img")

    # call TweakRegStep to update WCS & S_REGION
    res = trs.TweakRegStep.call([img])

    with res:
        for i, model in enumerate(res):
            assert model.meta.wcsinfo.s_region != old_fake_s_region
            res.shelve(model, i, modify=False)



"""Ramp Fitting tests involving MultiAccum Tables"""

import numpy as np
import pytest
from astropy import units as u
from astropy.time import Time
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import (
    DarkRefModel,
    GainRefModel,
    RampModel,
    ReadnoiseRefModel,
)

from romancal.ramp_fitting import RampFitStep

# Read Time in seconds
#   For Roman, the read time of the detectors is a fixed value and is currently
#   backed into code. Will need to refactor to consider the more general case.
#   Used to deconstruct the MultiAccum tables into integration times.
ROMAN_READ_TIME = 3.04

rng = np.random.default_rng(42)

# Basic resultant
#
# The read pattern is `[[1], [2], [3], [4]]`
# The total expected counts is 7.
# The resultants were generated with
# `romanisim.l1.apportion_counts_to_resultants(counts, read_pattern)`.
SIMPLE_RESULTANTS = np.array(
    [
        [[2.0, 2.0], [5.0, 1.0]],
        [[4.0, 5.0], [6.0, 2.0]],
        [[5.0, 6.0], [7.0, 6.0]],
        [[7.0, 7.0], [7.0, 7.0]],
    ],
    dtype=np.float32,
)
SIMPLE_EXPECTED_DEFAULT = {
    "data": np.array(
        [[0.52631587, 0.52631587], [0.23026317, 0.7236843]], dtype=np.float32
    ),
    "err": np.array(
        [[0.24262409, 0.24262409], [0.16048454, 0.28450054]], dtype=np.float32
    ),
    "var_poisson": np.array(
        [[0.05886428, 0.05886428], [0.02575312, 0.08093839]], dtype=np.float32
    ),
    "var_rnoise": np.array(
        [[2.164128e-06, 2.164128e-06], [2.164128e-06, 2.164128e-06]], dtype=np.float32
    ),
}
SIMPLE_EXPECTED_GAIN = {
    "data": np.array([[0.526316, 0.526316], [0.230263, 0.701852]], dtype=np.float32),
    "err": np.array([[0.108513, 0.108513], [0.071783, 0.124624]], dtype=np.float32),
    "var_poisson": np.array(
        [[1.1772858e-02, 1.1772858e-02], [5.150624e-03, 1.55289e-02]], dtype=np.float32
    ),
    "var_rnoise": np.array(
        [[2.164127e-06, 2.164127e-06], [2.164127e-06, 2.190606e-06]], dtype=np.float32
    ),
}
SIMPLE_EXPECTED_RNOISE = {
    "data": np.array(
        [[0.52631587, 0.52631587], [0.23026317, 0.7236843]], dtype=np.float32
    ),
    "err": np.array([[14.712976, 14.712976], [14.711851, 14.713726]], dtype=np.float32),
    "var_poisson": np.array(
        [[0.05886428, 0.05886428], [0.02575312, 0.08093839]], dtype=np.float32
    ),
    "var_rnoise": np.array(
        [[216.4128, 216.4128], [216.4128, 216.4128]], dtype=np.float32
    ),
}


# #####
# Tests
# #####
def test_bad_readpattern():
    """Ensure error is raised on bad readpattern"""
    ramp_model, gain_model, readnoise_model, dark_model = make_data(
        SIMPLE_RESULTANTS, 1, 0.01, False
    )
    bad_pattern = ramp_model.meta.exposure.read_pattern.data[1:]
    ramp_model.meta.exposure.read_pattern = bad_pattern

    with pytest.raises(RuntimeError):
        RampFitStep.call(
            ramp_model,
            algorithm="ols_cas22",
            override_gain=gain_model,
            override_readnoise=readnoise_model,
            override_dark=dark_model,
        )


@pytest.mark.parametrize(
    "attribute",
    ["data", "err", "var_poisson", "var_rnoise"],
    ids=["data", "err", "var_poisson", "var_rnoise"],
)
def test_fits(fit_ramps, attribute):
    """Check slopes"""
    image_model, expected = fit_ramps

    value = getattr(image_model, attribute)
    np.testing.assert_allclose(value, expected[attribute], 1e-05)


# ########
# Fixtures
# ########
@pytest.fixture(
    scope="module",
    params=[
        pytest.param(
            (SIMPLE_RESULTANTS, 1, 0.01, False, SIMPLE_EXPECTED_DEFAULT, False),
            id="default",
        ),  # No gain or noise
        pytest.param(
            (SIMPLE_RESULTANTS, 5, 0.01, False, SIMPLE_EXPECTED_GAIN, False),
            id="extragain",
        ),  # Increase gain
        pytest.param(
            (SIMPLE_RESULTANTS, 1, 100.0, False, SIMPLE_EXPECTED_RNOISE, False),
            id="extranoise",
        ),  # Increase noise
        pytest.param(
            (SIMPLE_RESULTANTS, 1, 0.01, False, SIMPLE_EXPECTED_DEFAULT, True),
            id="default-jump",
        ),  # No gain or noise
        pytest.param(
            (SIMPLE_RESULTANTS, 5, 0.01, False, SIMPLE_EXPECTED_GAIN, True),
            id="extragain-jump",
        ),  # Increase gain
        pytest.param(
            (SIMPLE_RESULTANTS, 1, 100.0, False, SIMPLE_EXPECTED_RNOISE, True),
            id="extranoise-jump",
        ),  # Increase noise
    ],
)
def fit_ramps(request):
    """Test ramp fits"""
    resultants, ingain, rnoise, randomize, expected, use_jump = request.param
    ramp_model, gain_model, readnoise_model, dark_model = make_data(
        resultants, ingain, rnoise, randomize
    )

    out_model = RampFitStep.call(
        ramp_model,
        algorithm="ols_cas22",
        use_ramp_jump_detection=use_jump,
        override_gain=gain_model,
        override_readnoise=readnoise_model,
        override_dark=dark_model,
    )

    return out_model, expected


# #########
# Utilities
# #########
def make_data(resultants, ingain, rnoise, randomize):
    """Create test input data

    Parameters
    ----------
    resultants : numpy.array.shape(3, xdim, ydim)
        The resultant array.

    ingain : int
        Gain to apply

    rnoise : float
        Noise to apply

    randomize : bool, expected)
        Randomize the gain and read noise across pixels.

    Returns
    -------
    image, gain, readnoise : ImageModel, GainRefModel, ReadnoiseRefModel.array
        Input image and related references
    """
    ramp_model = model_from_resultants(resultants)
    gain_model, readnoise_model, dark_model = generate_wfi_reffiles(
        ramp_model.shape[1:], ingain=ingain, rnoise=rnoise, randomize=randomize
    )

    return ramp_model, gain_model, readnoise_model, dark_model


def model_from_resultants(resultants, read_pattern=None):
    """Create a RampModel from resultants

    Parameters
    ----------
    resultants : numpy.array.shape(reads, xdim, ydim)
        The resultants to fit.

    read_pattern : [[int[,...]][,...]]
        The read pattern used to produce the resultants.
        If None, presume a basic read pattern
    """
    if read_pattern is None:
        read_pattern = [[idx + 1] for idx in range(resultants.shape[0])]

    # Full WFI image has reference pixels all around. Add those on.
    nrefpixs = 4
    full_wfi = np.ones(
        (
            resultants.shape[0],
            resultants.shape[1] + (nrefpixs * 2),
            resultants.shape[2] + (nrefpixs * 2),
        ),
        dtype=np.float32,
    )
    full_wfi[:, nrefpixs:-nrefpixs, nrefpixs:-nrefpixs] = resultants
    shape = full_wfi.shape

    pixdq = np.zeros(shape=shape[1:], dtype=np.uint32)
    err = np.zeros(shape=shape, dtype=np.float32)
    gdq = np.zeros(shape=shape, dtype=np.uint8)

    dm_ramp = maker_utils.mk_ramp(shape=shape)
    dm_ramp.data = full_wfi
    dm_ramp.pixeldq = pixdq
    dm_ramp.groupdq = gdq
    dm_ramp.err = err

    dm_ramp.meta.exposure.frame_time = ROMAN_READ_TIME
    dm_ramp.meta.exposure.nresultants = shape[0]

    dm_ramp.meta.exposure.read_pattern = read_pattern

    ramp_model = RampModel(dm_ramp)

    return ramp_model


def generate_wfi_reffiles(
    shape, ingain=6, rnoise=0.01, darkcurrent=0.01, randomize=True
):
    """Create GainRefModel and ReadnoiseRefModel

    Parameters
    ----------
    shape : tuple
        Shape of the arrays

    ingain : float
        Maximum gain.

    rnoise : flota
        Maximum noise

    darkcurrent : float
        Dark current scale

    randomize : bool
        Randomize the gain and read noise data.
    """
    # Create temporary gain reference file
    gain_ref = maker_utils.mk_gain(shape=shape)

    gain_ref["meta"]["instrument"]["detector"] = "WFI01"
    gain_ref["meta"]["instrument"]["name"] = "WFI"
    gain_ref["meta"]["reftype"] = "GAIN"
    gain_ref["meta"]["useafter"] = Time("2022-01-01T11:11:11.111")

    if randomize:
        gain_ref["data"] = u.Quantity(
            (rng.random(shape) * 0.5).astype(np.float32) * ingain,
            u.electron / u.DN,
            dtype=np.float32,
        )
    else:
        gain_ref["data"] = u.Quantity(
            np.ones(shape).astype(np.float32) * ingain,
            u.electron / u.DN,
            dtype=np.float32,
        )
    gain_ref["dq"] = np.zeros(shape, dtype=np.uint16)
    gain_ref["err"] = u.Quantity(
        (rng.random(shape) * 0.05).astype(np.float32),
        u.electron / u.DN,
        dtype=np.float32,
    )

    gain_ref_model = GainRefModel(gain_ref)

    # Create temporary readnoise reference file
    rn_ref = maker_utils.mk_readnoise(shape=shape)
    rn_ref["meta"]["instrument"]["detector"] = "WFI01"
    rn_ref["meta"]["instrument"]["name"] = "WFI"
    rn_ref["meta"]["reftype"] = "READNOISE"
    rn_ref["meta"]["useafter"] = Time("2022-01-01T11:11:11.111")

    rn_ref["meta"]["exposure"]["type"] = "WFI_IMAGE"
    rn_ref["meta"]["exposure"]["frame_time"] = 666

    if randomize:
        rn_ref["data"] = u.Quantity(
            (rng.random(shape) * rnoise).astype(np.float32),
            u.DN,
            dtype=np.float32,
        )
    else:
        rn_ref["data"] = u.Quantity(
            np.ones(shape).astype(np.float32) * rnoise, u.DN, dtype=np.float32
        )

    rn_ref_model = ReadnoiseRefModel(rn_ref)

    # Create temporary dark reference file
    # shape needs to be 3D but does not matter because the ramp fitting
    # step only uses the 2-D dark slope component
    dark_ref = maker_utils.mk_dark(shape=(1,) + shape)
    dark_ref["meta"]["instrument"]["detector"] = "WFI01"
    dark_ref["meta"]["instrument"]["name"] = "WFI"
    dark_ref["meta"]["reftype"] = "DARK"
    dark_ref["meta"]["useafter"] = Time("2022-01-01T11:11:11.111")

    dark_ref["meta"]["exposure"]["type"] = "WFI_IMAGE"
    dark_ref["meta"]["exposure"]["frame_time"] = 666

    dark_ref["dark_slope"] = u.Quantity(
        np.zeros(shape).astype(np.float32) * 0.01, u.DN / u.s, dtype=np.float32
    )

    dark_ref_model = DarkRefModel(dark_ref)

    # return gainfile, readnoisefile
    return gain_ref_model, rn_ref_model, dark_ref_model



"""Test suffix replacement"""

import logging

import pytest

from romancal.lib import suffix as s
from romancal.lib.basic_utils import LoggingContext


@pytest.fixture(scope="module")
def enable_logging():
    with LoggingContext(logging.getLogger(s.__name__), handler=logging.StreamHandler()):
        yield


@pytest.mark.filterwarnings("ignore:romancal.dqflags is deprecated.*")
def test_suffix_existence(enable_logging):
    """Generate current suffix list and compare"""

    calculated_suffixes = s.find_suffixes()
    found_suffixes = s.combine_suffixes(
        to_add=(calculated_suffixes, s.SUFFIXES_TO_ADD),
        to_remove=(s.SUFFIXES_TO_DISCARD,),
    )
    diff = set(found_suffixes).symmetric_difference(s.KNOW_SUFFIXES)
    assert not diff, f"Suffixes unaccounted for:\n{diff}"


@pytest.mark.parametrize("suffix", s.KNOW_SUFFIXES)
def test_suffix_removal(suffix, enable_logging):
    """Test suffix removal"""
    basename = "file"
    full_fpath = basename + "_" + suffix
    removed_path, separator = s.remove_suffix(full_fpath)
    assert removed_path == basename
    assert separator == "_"


@pytest.mark.parametrize("suffix", s.KNOW_SUFFIXES)
def test_suffix_replacement(suffix, enable_logging, base="file", new="junk", sep="_"):
    """Test suffix replacement"""
    full_path = base + sep + suffix
    replaced = s.replace_suffix(full_path, new)
    assert replaced == base + sep + new



"""Configure Association Logging"""

import logging
import sys
from collections import defaultdict
from logging.config import dictConfig

__all__ = ["log_config"]

DMS_DEFAULT_FORMAT = (
    "%(asctime)s %(levelname)s pid=%(process)d src=%(name)s.%(funcName)s"
)


class ContextFilter:
    """Set Association Generator logging context"""

    def __init__(self):
        self.context = {}

    def __call__(self):
        return self

    def filter(self, record):
        record._context = self.context
        return True

    def set(self, key, value):
        self.context[key] = value


class LogLevelFilter:
    """Filter on a specific level"""

    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno == self.__level


class DMSFormatter(logging.Formatter):
    """DMS-specific formatting"""

    def format(self, record):
        log_parts = []
        log_parts.append(super().format(record))
        for key in record._context:
            log_parts.append(f"{key}={record._context[key]}")
        log_parts.append(f'msg="{record.getMessage()}"')
        log_line = " ".join(log_parts)
        return log_line


# Define the common logging configuration
# Basic logger has the following config
base_logger = {
    "handlers": [
        "info",
        "debug",
        "default",
    ],
    "propagate": False,
}

# Basic, user-centric logging
context = ContextFilter()
base_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "datefmt": "%Y%m%d%H%M",
    "formatters": {
        "info": {"format": "%(message)s", "datefmt": "cfg://datefmt"},
        "debug": {
            "format": "%(asctime)s:%(levelname)s:%(name)s.%(funcName)s:%(message)s",
            "datefmt": "cfg://datefmt",
        },
    },
    "handlers": {
        "info": {
            "class": "logging.StreamHandler",
            "formatter": "info",
            "level": logging.INFO,
            "stream": sys.stdout,
            "filters": ["info", "context"],
        },
        "debug": {
            "class": "logging.StreamHandler",
            "formatter": "debug",
            "level": logging.DEBUG,
            "stream": sys.stderr,
            "filters": ["debug", "context"],
        },
        "default": {
            "class": "logging.StreamHandler",
            "formatter": "debug",
            "level": logging.WARN,
            "stream": sys.stderr,
            "filters": ["context"],
        },
    },
    "filters": {
        "context": {"()": context},
        "info": {"()": LogLevelFilter, "level": logging.INFO},
        "debug": {"()": LogLevelFilter, "level": logging.DEBUG},
    },
    "DMS": {
        "datefmt": "%Y%m%d%H%M",
        "logformat": (
            "%(asctime)s %(levelname)s pid=%(process)d src=%(name)s.%(funcName)s"
        ),
    },
}

# DMS-specific configuration
DMS_config = {
    "formatters": {
        "info": {
            "()": DMSFormatter,
            "format": "cfg://DMS.logformat",
            "datefmt": "cfg://DMS.datefmt",
        },
        "debug": {
            "()": DMSFormatter,
            "format": "cfg://DMS.logformat",
            "datefmt": "cfg://DMS.datefmt",
        },
    }
}


def log_config(name=None, user_name=None, logger_config=None, config=None, merge=True):
    """Setup logging with defaults

    logging.dictConfig is used with optional default
    configuration dict.

    Parameters
    ----------
    name : str
        The name of the logger to instantiate.
        If None, the root logger will be modified.

    user_name : str
        User-understandable name. If not specified, it will be the same
        as `name`.

    logger_config : dict
        The dict to use to setup the logger. This is used
        as the value to the key of `name`.

    config : dict
        User-specified logging configuration as specified by

    merge : boolean
        Merge the user-specified config in with `base_config`.
        If `False`, just use the user-specified config.


    Returns
    -------
    The `logging.Logger` instance

    Notes
    -----
    Internally, the logging configuration is placed into a dict defined by
    docs.python.org/3.5/library/logging.config.html#configuration-dictionary-schema
    `base_config` defines a set of handlers, formatters, and filters that
    separate out the different levels as follows:

        `default`
            Handles `logging.WARNING` and above.

        `debug`
            Handles `logging.DEBUG` exclusively.

        `info`
            Handles `logging.INFO` exclusively.

    Given the defaults, a logger is created with the given `name`,
    with the levels as described above.

    To make modifications, the rest of this note
    describes how the parameters interact
    with the configuration dict.

    The pair of parameters `name` and `logger_config` together make up the
    dict that is the value to the top-level key `loggers`

    The parameter `config` defines the rest of the top-level keys for
    configuration dict.

    When the parameter `merge` is `True`, the following happens:
    For the `formatters` dict, the dict for the key `name` is created by
    first creating the default dict, and then doing an
    `dict.update(logger_config)` to create the logger configuration.

    For the rest of the config, and dict.update is done on each of
    the top-level keys from the `config`.
    """

    global context

    if user_name is None:
        user_name = "root" if name is None else name
    if logger_config is None:
        logger_config = {}
    if config is None:
        config = {}
    use_logger = base_logger.copy()
    use_config = defaultdict(dict, base_config)
    if merge:
        use_logger.update(logger_config)
        for key in config:
            use_config[key].update(config[key])
        use_config["loggers"] = {name: use_logger}

    dictConfig(use_config)
    logger = logging.getLogger(name)
    logger.context = context
    return logger



#! /usr/bin/env python

import roman_datamodels as rdm
from roman_datamodels.datamodels import SaturationRefModel

from romancal.saturation import saturation
from romancal.stpipe import RomanStep

__all__ = ["SaturationStep"]


class SaturationStep(RomanStep):
    """
    This Step sets saturation flags.
    """

    class_alias = "saturation"

    reference_file_types = ["saturation"]

    def process(self, input):
        if isinstance(input, rdm.DataModel):
            input_model = input
        else:
            # Open the input data model
            input_model = rdm.open(input)

        # Get the name of the saturation reference file
        self.ref_name = self.get_reference_file(input_model, "saturation")

        # Check for a valid reference file
        if self.ref_name == "N/A":
            self.log.warning("No SATURATION reference file found")
            self.log.warning("Saturation step will be skipped")
            result = input_model.copy()
            result.meta.cal_step.saturation = "SKIPPED"
            return result

        # Open the reference file data model
        # Test for reference file
        self.log.info("Using SATURATION reference file: %s", self.ref_name)
        ref_model = SaturationRefModel(self.ref_name)

        # Perform saturation check
        sat = saturation.flag_saturation(input_model, ref_model)

        # Close the reference file and update the step status
        ref_model.close()
        sat.meta.cal_step.saturation = "COMPLETE"

        if self.save_results:
            try:
                self.suffix = "saturation"
            except AttributeError:
                self["suffix"] = "saturation"

        return sat



"""Test Level2 candidate operation"""

import pytest

from romancal.associations.main import Main
from romancal.associations.tests.helpers import (
    level2_rule_path,
    mkstemp_pool_file,
    t_path,
)


@pytest.mark.parametrize(
    "partial_args, n_asns",
    [
        # Invalid ACID
        (["-i", "nosuchid"], 0),
        # Basic observation ACIDs
        (["-i", "o001"], 0),
        # Whole program
        ([], 5),
        # Discovered only
        (["--discover"], 0),
        # Candidates only
        (["--all-candidates"], 5),
    ],
)
def test_candidate_observation(partial_args, n_asns):
    with mkstemp_pool_file(t_path("data/pool_001_candidates.csv")) as pool_path:
        cmd_args = [
            pool_path,
            "--dry-run",
            "-r",
            level2_rule_path(),
            "--ignore-default",
        ]
        cmd_args.extend(partial_args)
        generated = Main.cli(cmd_args)
        assert len(generated.associations) == n_asns



""" Roman tests for the High Level Pipeline """

import os

import pytest
import roman_datamodels as rdm

from romancal.pipeline.mosaic_pipeline import MosaicPipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(scope="module")
def run_mos(rtdata_module):
    rtdata = rtdata_module

    rtdata.get_asn("WFI/image/L3_regtest_asn.json")

    # Test Pipeline
    output = "r0099101001001001001_F158_visit_i2d.asdf"
    rtdata.output = output

    args = [
        "roman_mos",
        rtdata.input,
    ]
    MosaicPipeline.from_cmdline(args)

    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_mos):
    return run_mos.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_mos):
    return run_mos.truth


@pytest.fixture(scope="module")
def thumbnail_filename(output_filename):
    thumbnail_filename = output_filename.rsplit("_", 1)[0] + "_thumb.png"
    preview_cmd = f"stpreview to {output_filename} {thumbnail_filename} 256 256 roman"
    os.system(preview_cmd)  # nosec
    return thumbnail_filename


@pytest.fixture(scope="module")
def preview_filename(output_filename):
    preview_filename = output_filename.rsplit("_", 1)[0] + "_preview.png"
    preview_cmd = f"stpreview to {output_filename} {preview_filename} 1080 1080 roman"
    os.system(preview_cmd)  # nosec
    return preview_filename


def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    # DMS356
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_thumbnail_exists(thumbnail_filename):
    # DMS356
    assert os.path.isfile(thumbnail_filename)


def test_preview_exists(preview_filename):
    # DMS356
    assert os.path.isfile(preview_filename)


@pytest.mark.parametrize("suffix", ("cat", "segm"))
def test_file_exists(output_filename, suffix):
    # DMS374 for catalog and segm
    expected_filename = output_filename.rsplit("_", 1)[0] + f"_{suffix}.asdf"
    assert os.path.isfile(expected_filename)


def test_output_is_mosaic(output_model):
    # DMS356
    assert isinstance(output_model, rdm.datamodels.MosaicModel)


@pytest.mark.parametrize(
    "step_name",
    (
        "skymatch",
        "outlier_detection",
        "resample",
    ),
)
def test_steps_ran(output_model, step_name):
    # DMS356
    # DMS400 for skymatch
    # DMS86 for outlier_detection and resample
    assert getattr(output_model.meta.cal_step, step_name) == "COMPLETE"


def test_added_background(output_model):
    # DMS400
    assert hasattr(output_model.meta.individual_image_meta, "background")


def test_added_background_level(output_model):
    # DMS400
    assert any(output_model.meta.individual_image_meta.background["level"] != 0)



"""
Apply linearity correction to a science image
"""

import numpy as np
from roman_datamodels import datamodels as rdd
from roman_datamodels.dqflags import pixel
from stcal.linearity.linearity import linearity_correction

from romancal.stpipe import RomanStep

__all__ = ["LinearityStep"]


class LinearityStep(RomanStep):
    """
    LinearityStep: This step performs a correction for non-linear
    detector response, using the "classic" polynomial method.
    """

    class_alias = "linearity"

    reference_file_types = ["linearity"]

    def process(self, input):
        # Open the input data model
        if isinstance(input, rdd.DataModel):
            input_model = input
        else:
            input_model = rdd.open(input)

        # Get the name of the linearity reference file to use
        self.lin_name = self.get_reference_file(input_model, "linearity")
        self.log.info("Using LINEARITY reference file: %s", self.lin_name)

        # Check for a valid reference file
        if self.lin_name == "N/A":
            self.log.warning("No LINEARITY reference file found")
            self.log.warning("Linearity step will be skipped")
            input_model.meta.cal_step["linearity"] = "SKIPPED"

            return input_model

        with rdd.LinearityRefModel(self.lin_name, memmap=False) as lin_model:
            # copy poly coeffs from linearity model so Nan's can be updated
            lin_coeffs = lin_model.coeffs
            lin_dq = lin_model.dq  # 2D pixeldq from linearity model

            gdq = input_model.groupdq  # groupdq array of input model
            pdq = input_model.pixeldq  # pixeldq array of input model

            gdq = gdq[np.newaxis, :]

            input_model.data = input_model.data[np.newaxis, :]

            # Call linearity correction function in stcal
            # The third return value is the procesed zero frame which
            # Roman does not use.
            new_data, new_pdq, _ = linearity_correction(
                input_model.data, gdq, pdq, lin_coeffs, lin_dq, pixel
            )

            input_model.data = new_data[0, :, :, :]
            input_model.pixeldq = new_pdq

        # Update the step status
        input_model.meta.cal_step["linearity"] = "COMPLETE"

        if self.save_results:
            try:
                self.suffix = "linearity"
            except AttributeError:
                self["suffix"] = "linearity"
        return input_model



"""
Unit tests for patch_match.

These tests depend very strongly on the contents of the referenced table of patches.
Changes to the contents of this table will require changes to the tests for
any unit tests that depend on specific matches to patches in the table.

Any changes to the matching algorithm should be completely separate from any
changes to the table contents.

The tests include the following cases to validate the matches
1) Simple case at ra, dec, pa = 0, 0, 0
2) 1) translated to ra, de, pa = 180, 40, 0
3) Same as 2) but with pa = 45 and 60.
4) Same as 3) (pa=45 case) but with the lower corner just above, below, to the right
   and to the left of a 4 patch corner
   (assuming non-overlapping patchs within a common tangent point).
   This requires identifying the ra, dec of such a corner in the table.
5) A test of a WCS provided example.

Most of these tests check to see if the matches are what are expected, by index of
the table (the table format is expected to include the index of the entry as one
of its columns so that subsets of the table selected to reduce the filesize still
retain the same index obtained.)
"""

from pathlib import Path

import asdf
import astropy.coordinates as coord
import astropy.modeling.models as amm
import astropy.units as u
import gwcs.coordinate_frames as cf
import gwcs.wcs as wcs
import numpy as np
import pytest
import spherical_geometry.vector as sgv
from spherical_geometry.vector import rotate_around as rotate

import romancal.patch_match.patch_match as pm

PATCH_SUBSET_PATH = Path(__file__).parent / "patches_subset.asdf"
# do not use load_patch_table here as it will modify global state
with asdf.open(PATCH_SUBSET_PATH) as af:
    PATCH_SUBSET = af["patches"].copy()

crecord = PATCH_SUBSET[np.where(PATCH_SUBSET[:]["index"] == 925050)]
cra = crecord["ra_corn3"]
if len(cra) == 1:
    cra = cra[0]
cdec = crecord["dec_corn3"]
if len(cdec) == 1:
    cdec = cdec[0]

cpa = 45.0
csize = 0.001
e = 0.0011  # epsilon offset in degrees


@pytest.fixture(autouse=True)
def override_patch_table(monkeypatch):
    """
    For the tests in this file, monkeypatch the global
    PATCH_TABLE to a smaller PATCH_SUBSET to allow these tests
    to run without access to the full patch table.
    """
    monkeypatch.setattr(pm, "PATCH_TABLE", PATCH_SUBSET)
    yield


def mk_im_corners(ra, dec, pa, size):
    """
    Generate 4 image corners of a square with the center at the supplied
    side size, ra, dec, and position angle (all in degrees).
    """
    # Generate 4 unit vectors at ra, dec = (0 , 0)
    center = sgv.lonlat_to_vector(0.0, 0.0)
    radecvec = sgv.lonlat_to_vector(ra, dec)
    zaxis = (0.0, 0.0, 1.0)
    yaxis = (0.0, 1.0, 0.0)
    pp = rotate(*(rotate(*(center + yaxis + (-size / 2,))) + zaxis + (+size / 2,)))
    pm = rotate(*(rotate(*(center + yaxis + (+size / 2,))) + zaxis + (+size / 2,)))
    mp = rotate(*(rotate(*(center + yaxis + (-size / 2,))) + zaxis + (-size / 2,)))
    mm = rotate(*(rotate(*(center + yaxis + (+size / 2,))) + zaxis + (-size / 2,)))
    rect = [pp, mp, mm, pm]

    # Now move to requested ra and dec
    trect = [
        rotate(*(rotate(*(vec + yaxis + (-dec,))) + zaxis + (ra,))) for vec in rect
    ]
    # Rotate to desired position angle
    rrect = [rotate(*(vec + radecvec + (pa,))) for vec in trect]
    frect = [sgv.vector_to_lonlat(*vec) for vec in rrect]
    # Reorganize by ra, dec arrays
    radecrect = np.array(frect).transpose()
    return radecrect


def mk_gwcs(ra=cra, dec=cdec, pa=cpa, bounding_box=None, pixel_shape=None):
    """
    Construct a GWCS model for testing the patch matching when provided a WCS
    This just implements a basic tangent projection with specified ra, dec, and
    position angle
    """
    transform = (amm.Shift(-2048) & amm.Shift(-2048)) | (
        amm.Scale(0.11 / 3600.0) & amm.Scale(0.11 / 3600.0)
        | amm.Rotation2D(pa)
        | amm.Pix2Sky_TAN()
        | amm.RotateNative2Celestial(cra, cdec, 180.0)
    )
    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    wcsobj = wcs.WCS([(detector_frame, transform), (sky_frame, None)])
    if pixel_shape is not None:
        wcsobj.pixel_shape = pixel_shape
    if bounding_box is not None:
        wcsobj.bounding_box = bounding_box
    return wcsobj


@pytest.mark.parametrize(
    "pars, expected",
    [
        ((cra, cdec + e, cpa, csize), (925051, 925151)),
        ((cra, cdec - e, cpa, csize), (925050, 925150)),
        ((cra + e, cdec, cpa, csize), (925150, 925151)),
        ((cra - e, cdec, cpa, csize), (925050, 925051)),
        ((cra, cdec, cpa, csize), (925050, 925051, 925150, 925151)),
        (
            (cra, cdec, cpa, 0.5),
            (
                924750,
                924751,
                924849,
                924850,
                924851,
                924852,
                924948,
                924949,
                924950,
                924951,
                924952,
                924953,
                925047,
                925048,
                925049,
                925050,
                925051,
                925052,
                925053,
                925054,
                925147,
                925148,
                925149,
                925150,
                925151,
                925152,
                925153,
                925154,
                925248,
                925249,
                925250,
                925251,
                925252,
                925253,
                925349,
                925350,
                925351,
                925352,
                925450,
                925451,
            ),
        ),
    ],
)
def test_corners(pars, expected):
    corners = mk_im_corners(*pars)
    matches, close = pm.find_patch_matches(corners)
    # map matches to absolute index
    mmatches = tuple([PATCH_SUBSET[match]["index"] for match in matches])
    assert tuple(mmatches) == expected


def test_wcs_corners():
    imshape = (4096, 4096)
    wcsobj = mk_gwcs()
    matches, close = pm.find_patch_matches(wcsobj, image_shape=imshape)
    mmatches = tuple([PATCH_SUBSET[match]["index"] for match in matches])
    assert tuple(mmatches) == (925050, 925051, 925150, 925151)
    wcsobj.pixel_shape = imshape
    matches, close = pm.find_patch_matches(wcsobj)
    mmatches = tuple([PATCH_SUBSET[match]["index"] for match in matches])
    assert tuple(mmatches) == (925050, 925051, 925150, 925151)
    wcsobj.pixel_shape = None
    wcsobj.bounding_box = ((-0.5, 4096 - 0.5), (-0.5, 4096 - 0.5))
    matches, close = pm.find_patch_matches(wcsobj)
    mmatches = tuple([PATCH_SUBSET[match]["index"] for match in matches])
    assert tuple(mmatches) == (925050, 925051, 925150, 925151)
    wcsobj.bounding_box = None
    with pytest.raises(ValueError):
        matches, close = pm.find_patch_matches(wcsobj)



import os

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
from roman_datamodels.datamodels import MosaicModel
from roman_datamodels.maker_utils import mk_level3_mosaic

from romancal.datamodels import ModelLibrary
from romancal.multiband_catalog import MultibandCatalogStep


def make_test_image():
    g1 = Gaussian2D(121.0, 11.1, 12.2, 1.5, 1.5)
    g2 = Gaussian2D(70, 65, 18, 9.2, 4.5)
    g3 = Gaussian2D(111.0, 41, 42.7, 8.0, 3.0, theta=30 * u.deg)
    g4 = Gaussian2D(81.0, 17, 52.7, 4, 2, theta=102 * u.deg)
    g5 = Gaussian2D(107.0, 65, 71, 12, 2, theta=142 * u.deg)
    g6 = Gaussian2D(50, 20, 80, 2.1, 2.1)
    g7 = Gaussian2D(97.0, 85, 87.3, 4, 2, theta=-30 * u.deg)

    yy, xx = np.mgrid[0:101, 0:101]
    data = (
        g1(xx, yy)
        + g2(xx, yy)
        + g3(xx, yy)
        + g4(xx, yy)
        + g5(xx, yy)
        + g6(xx, yy)
        + g7(xx, yy)
    ).value.astype("float32")

    rng = np.random.default_rng(seed=123)
    noise = rng.normal(0, 2.5, size=data.shape)
    data += noise
    err = data / 10.0

    return data, err


@pytest.fixture
def mosaic_model():
    wfi_mosaic = mk_level3_mosaic(shape=(101, 101))
    model = MosaicModel(wfi_mosaic)
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.var_rnoise = err**2
    model.weight = 1.0 / err
    return model


@pytest.fixture
def library_model(mosaic_model):
    model2 = mosaic_model.copy()
    model2.meta.basic.optical_element = "F184"
    return ModelLibrary([mosaic_model, model2])


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "snr_threshold, npixels, save_results",
    (
        (3, 10, True),
        (10, 10, False),
    ),
)
def test_multiband_catalog(
    library_model, snr_threshold, npixels, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = MultibandCatalogStep()

    result = step.call(
        library_model,
        bkg_boxsize=50,
        snr_threshold=snr_threshold,
        npixels=npixels,
        fit_psf=False,
        save_results=save_results,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)

    columns = [
        "label",
        "xcentroid",
        "ycentroid",
        "ra_centroid",
        "dec_centroid",
        "is_extended",
        "sharpness",
        "roundness",
        "nn_label",
        "nn_dist",
        "isophotal_area",
        "semimajor_sigma",
        "semiminor_sigma",
        "ellipticity",
        "orientation",
        "sky_orientation",
        "ra_bbox_ll",
        "dec_bbox_ll",
        "ra_bbox_ul",
        "dec_bbox_ul",
        "ra_bbox_lr",
        "dec_bbox_lr",
        "ra_bbox_ur",
        "dec_bbox_ur",
    ]

    assert len(cat) == 7

    if len(cat) > 0:
        for col in columns:
            assert col in cat.colnames
        assert np.min(cat["xcentroid"]) > 0.0
        assert np.min(cat["ycentroid"]) > 0.0
        assert np.max(cat["xcentroid"]) < 100.0
        assert np.max(cat["ycentroid"]) < 100.0

        for phottype in ("isophotal", "kron", "aper30", "aper_total"):
            for filt in ("F158", "F184"):
                colname = f"{filt}_{phottype}_flux"
                assert colname in cat.colnames
                assert colname + "_err" in cat.colnames



import asdf
import numpy as np
import pytest
from astropy import units as u
from roman_datamodels import datamodels as rdm

from romancal.stpipe import RomanStep
from romancal.tweakreg.tweakreg_step import TweakRegStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_tweakreg(rtdata, ignore_asdf_paths, tmp_path):
    # N.B.: uncal file is from simulator
    # ``shifted'' version is created in make_regtestdata.sh; cal file is taken,
    # the wcsinfo is perturbed, and AssignWCS is run to update the WCS with the
    # perturbed information
    orig_uncal = "r0000101001001001001_0001_wfi01_uncal.asdf"
    orig_catfile = "r0000101001001001001_0001_wfi01_cat.asdf"
    input_data = "r0000101001001001001_0001_wfi01_shift_cal.asdf"
    output_data = "r0000101001001001001_0001_wfi01_shift_tweakregstep.asdf"
    truth_data = "r0000101001001001001_0001_wfi01_shift_tweakregstep.asdf"

    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.get_data(f"WFI/image/{orig_catfile}")
    rtdata.get_data(f"WFI/image/{orig_uncal}")
    rtdata.get_truth(f"truth/WFI/image/{truth_data}")

    rtdata.input = input_data
    rtdata.output = output_data

    # instantiate TweakRegStep (for running and log access)
    step = TweakRegStep()

    args = [
        "romancal.step.TweakRegStep",
        rtdata.input,
        f"--output_file='{rtdata.output}'",
        "--suffix='tweakregstep'",
    ]
    RomanStep.from_cmdline(args)
    tweakreg_out = rdm.open(rtdata.output)

    step.log.info(
        "DMS280 MSG: TweakReg step recorded as complete? :"
        f' {tweakreg_out.meta.cal_step.tweakreg == "COMPLETE"}'
    )
    assert tweakreg_out.meta.cal_step.tweakreg == "COMPLETE"

    step.log.info(
        f"""DMS280 MSG: TweakReg created new attribute with fit results? :\
            {"wcs_fit_results" in tweakreg_out.meta}"""
    )
    assert "wcs_fit_results" in tweakreg_out.meta

    step.log.info(
        f"""DMS280 MSG: TweakReg created new coordinate frame 'v2v3corr'? :\
            {"v2v3corr" in tweakreg_out.meta.wcs.available_frames}"""
    )
    assert "v2v3corr" in tweakreg_out.meta.wcs.available_frames

    wcstweak = tweakreg_out.meta.wcs
    with asdf.open(orig_uncal) as orig_model_asdf:
        wcstrue = orig_model_asdf["romanisim"]["wcs"]  # simulated, true WCS
        pts = np.linspace(0, 4000, 30)
        xx, yy = np.meshgrid(pts, pts)
        coordtweak = wcstweak.pixel_to_world(xx, yy)
        coordtrue = wcstrue.pixel_to_world(xx, yy)
    diff = coordtrue.separation(coordtweak).to(u.arcsec).value
    rms = np.sqrt(np.mean(diff**2)) * 1000  # rms difference in mas
    passmsg = "PASS" if rms < 1.3 / np.sqrt(2) else "FAIL"
    step.log.info(
        f"DMS488 MSG: WCS agrees with true WCS to {rms:5.2f} mas, less than "
        f"1.3 / sqrt(2)?  {passmsg}"
    )
    step.log.info(
        f"DMS405 MSG: WCS agrees with true WCS to {rms:5.2f} mas, less than "
        f"5 / sqrt(2)?  {passmsg}"
    )

    assert rms < 1.3 / np.sqrt(2) * 100

    diff = compare_asdf(rtdata.output, rtdata.truth, atol=1e-3, **ignore_asdf_paths)
    step.log.info(
        f"DMS280 MSG: Was the proper TweakReg data produced? : {diff.identical}"
    )
    assert diff.identical, diff.report()
